# Bob build tool
# Copyright (C) 2016  Jan Klötzke
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ..errors import BobError
import argparse
import os.path
import subprocess
import sys

def doHelp(availableCommands, argv, bobRoot):
    parser = argparse.ArgumentParser(prog="bob help",
        description="Display help information about command.")
    # Help without a command parameter gets handled by the main argument parser
    # in pym/bob/scripts.py.
    parser.add_argument('command', help="Command to get help for")

    args = parser.parse_args(argv)

    if args.command in availableCommands:
        manPage = "bob-" + args.command
        manSection = "1"
    else:
        manPage = "bob" + args.command
        manSection = "7"

    try:
        from ..develop.make import makeManpages
        manPath = makeManpages()
        manArgs = [ os.path.join(manPath, manPage+"."+manSection) ]
    except ImportError:
        manArgs = [manSection, manPage]
    except OSError as e:
        raise BobError("Cannot build manpage: " + str(e))

    try:
        ret = subprocess.call(["man"] + manArgs)
    except OSError:
        ret = 1

    sys.exit(ret)



# Bob build tool
#
# SPDX-License-Identifier: GPL-3.0-or-later

# Generator for eclipse cdt project files

import argparse
import sys
import os
import re
import shutil
import stat
import uuid
from os.path import expanduser
from os.path import join
from bob.utils import summonMagic, removePath
from collections import OrderedDict
from shlex import quote

# scan package recursivelely with its dependencies and build a list of checkout dirs
def getCheckOutDirs(package, dirs):
    def collect(package, dirs, steps, processed):
        if package._getId() in processed:
            return
        processed.add(package._getId())
        if package.getCheckoutStep().isValid():
            if package.getCheckoutStep().getVariantId() not in steps:
                steps.add(package.getCheckoutStep().getVariantId())
                dirs.append([package.getName(), package.getCheckoutStep().getWorkspacePath()])
        for d in package.getDirectDepSteps():
            collect(d.getPackage(), dirs, steps, processed)

    collect(package, dirs, set(), set())

# generate a unique id
def getId():
    _uuid = uuid.uuid4()
    return str(int(_uuid.hex[0:7], 16))

def generateFile(entries, fileName):
    try:
        os.remove(fileName)
    except OSError:
        pass
    fileName = open(fileName, "w")
    for e in entries:
        fileName.write(e + "\n")
    fileName.close()

def createLaunchFile(outFile, exe, project):
    outFile.write('<?xml version="1.0" encoding="UTF-8" standalone="no"?>\n')
    outFile.write('<launchConfiguration type="org.eclipse.cdt.launch.applicationLaunchType">\n')
    outFile.write('<booleanAttribute key="org.eclipse.cdt.dsf.gdb.AUTO_SOLIB" value="true"/>\n')
    outFile.write('<listAttribute key="org.eclipse.cdt.dsf.gdb.AUTO_SOLIB_LIST"/>\n')
    outFile.write('<stringAttribute key="org.eclipse.cdt.dsf.gdb.DEBUG_NAME" value="gdb"/>\n')
    outFile.write('<booleanAttribute key="org.eclipse.cdt.dsf.gdb.DEBUG_ON_FORK" value="false"/>\n')
    outFile.write('<stringAttribute key="org.eclipse.cdt.dsf.gdb.GDB_INIT" value=".gdbinit"/>\n')
    outFile.write('<booleanAttribute key="org.eclipse.cdt.dsf.gdb.NON_STOP" value="false"/>\n')
    outFile.write('<booleanAttribute key="org.eclipse.cdt.dsf.gdb.REVERSE" value="false"/>\n')
    outFile.write('<listAttribute key="org.eclipse.cdt.dsf.gdb.SOLIB_PATH"/>\n')
    outFile.write('<stringAttribute key="org.eclipse.cdt.dsf.gdb.TRACEPOINT_MODE" value="TP_NORMAL_ONLY"/>\n')
    outFile.write('<booleanAttribute key="org.eclipse.cdt.dsf.gdb.UPDATE_THREADLIST_ON_SUSPEND" value="false"/>\n')
    outFile.write('<booleanAttribute key="org.eclipse.cdt.dsf.gdb.internal.ui.launching.LocalApplicationCDebuggerTab.DEFAULTS_SET" value="true"/>\n')
    outFile.write('<intAttribute key="org.eclipse.cdt.launch.ATTR_BUILD_BEFORE_LAUNCH_ATTR" value="2"/>\n')
    outFile.write('<stringAttribute key="org.eclipse.cdt.launch.COREFILE_PATH" value=""/>\n')
    outFile.write('<stringAttribute key="org.eclipse.cdt.launch.DEBUGGER_ID" value="gdb"/>\n')
    outFile.write('<stringAttribute key="org.eclipse.cdt.launch.DEBUGGER_REGISTER_GROUPS" value=""/>\n')
    outFile.write('<stringAttribute key="org.eclipse.cdt.launch.DEBUGGER_START_MODE" value="run"/>\n')
    outFile.write('<booleanAttribute key="org.eclipse.cdt.launch.DEBUGGER_STOP_AT_MAIN" value="true"/>\n')
    outFile.write('<stringAttribute key="org.eclipse.cdt.launch.DEBUGGER_STOP_AT_MAIN_SYMBOL" value="main"/>\n')
    outFile.write('<stringAttribute key="org.eclipse.cdt.launch.PROGRAM_NAME" value="'+ exe + '"/>\n')
    outFile.write('<stringAttribute key="org.eclipse.cdt.launch.PROJECT_ATTR" value="' + project + '"/>\n')
    outFile.write('<booleanAttribute key="org.eclipse.cdt.launch.PROJECT_BUILD_CONFIG_AUTO_ATTR" value="true"/>\n')
    outFile.write('<stringAttribute key="org.eclipse.cdt.launch.PROJECT_BUILD_CONFIG_ID_ATTR" value=""/>\n')
    outFile.write('<listAttribute key="org.eclipse.debug.core.MAPPED_RESOURCE_PATHS">\n')
    outFile.write('<listEntry value="/' + project + '"/>\n')
    outFile.write('</listAttribute>\n')
    outFile.write('<listAttribute key="org.eclipse.debug.core.MAPPED_RESOURCE_TYPES">\n')
    outFile.write('<listEntry value="4"/>\n')
    outFile.write('</listAttribute>\n')
    outFile.write('<listAttribute key="org.eclipse.debug.ui.favoriteGroups">\n')
    outFile.write('<listEntry value="org.eclipse.debug.ui.launchGroup.debug"/>\n')
    outFile.write('<listEntry value="org.eclipse.debug.ui.launchGroup.run"/>\n')
    outFile.write('</listAttribute>\n')
    outFile.write('<stringAttribute key="org.eclipse.dsf.launch.MEMORY_BLOCKS" value="&lt;?xml version=&quot;1.0&quot; encoding=&quot;UTF-8&quot; standalone=&quot;no&quot;?&gt;&#10;&lt;memoryBlockExpressionList context=&quot;reserved-for-future-use&quot;/&gt;&#10;"/>\n')
    outFile.write('<stringAttribute key="process_factory_id" value="org.eclipse.cdt.dsf.gdb.GdbProcessFactory"/>\n')
    outFile.write('</launchConfiguration>')

def addIncludes(cProjectFile, includeDirs):
    cProjectFile.write('<option id="org.eclipse.cdt.build.core.settings.holder.incpaths.'+getId()+'" name="Include Paths" superClass="org.eclipse.cdt.build.core.settings.holder.incpaths" valueType="includePath">\n')
    for i in includeDirs:
        cProjectFile.write(' <listOptionValue builtIn="false" value="'+i+'"/>\n')
    cProjectFile.write('</option>\n')

def addCConfig(cProjectFile, excludePackages, includeDirs, buildName, id, buildArgs, buildMeFile):
    cProjectFile.write('<cconfiguration id="' + id + '">\n')
    cProjectFile.write('<storageModule buildSystemId="org.eclipse.cdt.managedbuilder.core.configurationDataProvider" id="'
                        + id + '" moduleId="org.eclipse.cdt.core.settings" name="' + buildName + '">\n')
    cProjectFile.write(' <externalSettings/>\n')
    cProjectFile.write(' <extensions>\n')
    cProjectFile.write('  <extension id="org.eclipse.cdt.core.GASErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>\n')
    cProjectFile.write('  <extension id="org.eclipse.cdt.core.GmakeErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>\n')
    cProjectFile.write('  <extension id="org.eclipse.cdt.core.GLDErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>\n')
    cProjectFile.write('  <extension id="org.eclipse.cdt.core.VCErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>\n')
    cProjectFile.write('  <extension id="org.eclipse.cdt.core.CWDLocator" point="org.eclipse.cdt.core.ErrorParser"/>\n')
    cProjectFile.write('  <extension id="org.eclipse.cdt.core.GCCErrorParser" point="org.eclipse.cdt.core.ErrorParser"/>\n')
    cProjectFile.write(' </extensions>\n')
    cProjectFile.write('</storageModule>\n')
    cProjectFile.write('<storageModule moduleId="cdtBuildSystem" version="4.0.0">\n')
    cProjectFile.write(' <configuration buildProperties="" description="" id="' + id + '" name="' + buildName + '" parent="org.eclipse.cdt.build.core.prefbase.cfg">\n')
    cProjectFile.write('   <folderInfo id="' + id + '." name="/" resourcePath="">\n')
    toolchainId = getId()
    cProjectFile.write('    <toolChain id="org.eclipse.cdt.build.core.prefbase.toolchain.' + toolchainId + '" name="No ToolChain" resourceTypeBasedDiscovery="false" superClass="org.eclipse.cdt.build.core.prefbase.toolchain">\n')
    cProjectFile.write('     <targetPlatform id="org.eclipse.cdt.build.core.prefbase.toolchain.' + toolchainId + '.' + getId() + '" name=""/>\n')
    cProjectFile.write('     <builder command="' + buildMeFile + '" id="org.eclipse.cdt.build.core.settings.default.builder.' + getId()
            + '" incrementalBuildTarget="' + buildArgs
            + '" buildPath="' + os.getcwd()
            + '" keepEnvironmentInBuildfile="false" managedBuildOn="false" name="Gnu Make Builder" \
                    superClass="org.eclipse.cdt.build.core.settings.default.builder.' + getId() + '"/>\n')
    cProjectFile.write('     <tool id="org.eclipse.cdt.build.core.settings.holder.libs.' + getId() + '" name="holder for library settings" superClass="org.eclipse.cdt.build.core.settings.holder.libs"/>\n')
    cProjectFile.write('     <tool id="org.eclipse.cdt.build.core.settings.holder.' + getId() + '" name="Assembly" superClass="org.eclipse.cdt.build.core.settings.holder">\n')
    if len(includeDirs):
        addIncludes(cProjectFile, includeDirs)
    cProjectFile.write('             <inputType id="org.eclipse.cdt.build.core.settings.holder.inType.' + getId() + '" languageId="org.eclipse.cdt.core.assembly" languageName="Assembly" sourceContentType="org.eclipse.cdt.core.asmSource" superClass="org.eclipse.cdt.build.core.settings.holder.inType"/>\n')
    cProjectFile.write('     </tool>\n')
    cProjectFile.write('     <tool id="org.eclipse.cdt.build.core.settings.holder.' + getId() + '" name="GNU C++" superClass="org.eclipse.cdt.build.core.settings.holder">\n')
    if len(includeDirs):
        addIncludes(cProjectFile, includeDirs)
    cProjectFile.write('             <inputType id="org.eclipse.cdt.build.core.settings.holder.inType.' + getId() + '" languageId="org.eclipse.cdt.core.g++" languageName="GNU C++" sourceContentType="org.eclipse.cdt.core.cxxSource,org.eclipse.cdt.core.cxxHeader" superClass="org.eclipse.cdt.build.core.settings.holder.inType"/>\n')
    cProjectFile.write('     </tool>\n')
    cProjectFile.write('     <tool id="org.eclipse.cdt.build.core.settings.holder.' + getId() + '" name="GNU C" superClass="org.eclipse.cdt.build.core.settings.holder">\n')
    if len(includeDirs):
        addIncludes(cProjectFile, includeDirs)
    cProjectFile.write('             <inputType id="org.eclipse.cdt.build.core.settings.holder.inType.' + getId() + '" languageId="org.eclipse.cdt.core.gcc" languageName="GNU C" sourceContentType="org.eclipse.cdt.core.cSource,org.eclipse.cdt.core.cHeader" superClass="org.eclipse.cdt.build.core.settings.holder.inType"/>\n')
    cProjectFile.write('     </tool>\n')
    cProjectFile.write('    </toolChain>\n')
    cProjectFile.write('   </folderInfo>\n')

    cProjectFile.write('   <sourceEntries>\n')
    if len(excludePackages) > 0:
        excludes = ''
        for package in excludePackages:
            excludes += package + '|'
        cProjectFile.write('<entry excluding="' + excludes[:-1] + '" flags="VALUE_WORKSPACE_PATH|RESOLVED" kind="sourcePath" name=""/>\n')
    cProjectFile.write('   </sourceEntries>\n')

    cProjectFile.write(' </configuration>\n')
    cProjectFile.write('</storageModule>\n')
    cProjectFile.write('<storageModule moduleId="org.eclipse.cdt.core.externalSettings"/>\n')
    cProjectFile.write('</cconfiguration>\n')

def eclipseCdtGenerator(package, argv, extra, bobRoot):
    parser = argparse.ArgumentParser(prog="bob project eclipse-cdt", description='Generate Eclipse CDT Project Files')
    parser.add_argument('-u', '--update', default=False, action='store_true',
                        help="Update project files (.project)")
    parser.add_argument('--buildCfg', action='append', default=[], type=lambda a: a.split("::"),
         help="Adds a new buildconfiguration. Format: <Name>::<flags>")
    parser.add_argument('--overwrite', action='store_true',
        help="Remove destination folder before generating.")
    parser.add_argument('--destination', metavar="DEST",
        help="Destination of project files")
    parser.add_argument('--name', metavar="NAME",
        help="Name of project. Default is complete_path_to_package")
    parser.add_argument('--exclude', default=[], action='append', dest="excludes",
            help="Packages will be marked as 'exclude from build' in eclipse. Usefull if indexer runs OOM.")
    parser.add_argument('-I', dest="additional_includes", default=[], action='append',
        help="Additional include directories. (added recursive starting from this directory)")

    args = parser.parse_args(argv)
    extra = " ".join(quote(e) for e in extra)

    destination = args.destination
    projectName = args.name

    project = "/".join(package.getStack())

    dirs = []
    getCheckOutDirs(package, dirs)
    if not projectName:
        # use package name for project name
        projectName = package.getName()

    projectName = projectName.replace(':', '_')
    if not destination:
        # use package name for project directory
        destination = os.path.join(os.getcwd(), "projects", projectName)
    destination = destination.replace(':', '_')
    if args.overwrite:
        removePath(destination)
    if not os.path.exists(destination):
        os.makedirs(destination)

    buildMeFile = os.path.join(destination, "buildme")

    id = "0." + getId()

    cProjectHeader = '<?xml version="1.0" encoding="UTF-8" standalone="no"?>'
    cProjectHeader += '<?fileVersion 4.0.0?><cproject storage_type_id="org.eclipse.cdt.core.XmlProjectDescriptionStorage">'
    cProjectHeader += '<storageModule moduleId="org.eclipse.cdt.core.settings">'

    cProjectFooter  = '</storageModule>\n'
    cProjectFooter += '<storageModule moduleId="cdtBuildSystem" version="4.0.0">\n'
    cProjectFooter += ' <project id="bob4eclipse.null.' + getId() + '" name="bob4eclipse"/>\n'
    cProjectFooter += '</storageModule>\n'
    cProjectFooter += '<storageModule moduleId="scannerConfiguration">\n'
    cProjectFooter += ' <autodiscovery enabled="true" problemReportingEnabled="true" selectedProfileId=""/>\n'
    cProjectFooter += ' <scannerConfigBuildInfo instanceId=" ' + id + '">\n'
    cProjectFooter += '  <autodiscovery enabled="true" problemReportingEnabled="true" selectedProfileId=""/>\n'
    cProjectFooter += ' </scannerConfigBuildInfo>\n'
    cProjectFooter += '</storageModule>\n'
    cProjectFooter += '<storageModule moduleId="org.eclipse.cdt.core.LanguageSettingsProviders"/>\n'
    cProjectFooter += '<storageModule moduleId="refreshScope" versionNumber="2">\n'
    cProjectFooter += ' <configuration configurationName="bob (build only)">\n'
    cProjectFooter += '  <resource resourceType="PROJECT" workspacePath="/bob4eclipse"/>\n'
    cProjectFooter += ' </configuration>\n'
    cProjectFooter += ' <configuration configurationName="Default">\n'
    cProjectFooter += '  <resource resourceType="PROJECT" workspacePath="/bob4eclipse"/>\n'
    cProjectFooter += ' </configuration>\n'
    cProjectFooter += '</storageModule>\n'
    cProjectFooter += '</cproject>\n'

    # setup list of exclude packages
    excludePackages = []
    try:
        for e in args.excludes:
            exp = re.compile(e)
            for name,path in OrderedDict(sorted(dirs, key=lambda t: t[1])).items():
                if exp.search(name):
                    excludePackages.append(name)
        includeDirs = []
        # find additional include dirs
        for i in args.additional_includes:
            if os.path.exists(i):
                for root, directories, filenames in os.walk(i):
                    includeDirs.append(os.path.join(i,root))
    except re.error as e:
        raise ParseError("Invalid regular expression '{}': {}".format(e.pattern), e)

    with open(os.path.join(destination, ".cproject"), 'w') as cProjectFile:
        cProjectFile.write(cProjectHeader)
        addCConfig(cProjectFile, excludePackages, includeDirs, "Bob dev", id, "", buildMeFile)
        addCConfig(cProjectFile, excludePackages, includeDirs, "Bob dev (force)", id + "." + getId(), "-f", buildMeFile)
        addCConfig(cProjectFile, excludePackages, includeDirs, "Bob dev (no checkout)",id + "." + getId(), "-b", buildMeFile)
        addCConfig(cProjectFile, excludePackages, includeDirs, "Bob dev (no deps)",id + "." + getId(), "-n", buildMeFile)
        addCConfig(cProjectFile, excludePackages, includeDirs, "Bob dev (no checkout, no deps)",id + "." + getId(), "-bn", buildMeFile)
        addCConfig(cProjectFile, excludePackages, includeDirs, "Bob dev (no checkout, clean)",id + "." + getId(), "-b --clean", buildMeFile)

        for name,flags in args.buildCfg:
            addCConfig(cProjectFile, excludePackages, includeDirs, name, id + "." + getId(), flags, buildMeFile)

        cProjectFile.write(cProjectFooter)

    projectFileHeader = """<?xml version="1.0" encoding="UTF-8"?>
<projectDescription>
<name>""" + projectName + """</name>
<comment></comment>
<projects>
</projects>
<buildSpec>
 <buildCommand>
  <name>org.eclipse.cdt.managedbuilder.core.genmakebuilder</name>
  <arguments>
  </arguments>
 </buildCommand>
 <buildCommand>
  <name>org.eclipse.cdt.managedbuilder.core.ScannerConfigBuilder</name>
  <triggers>full,incremental,</triggers>
  <arguments>
  </arguments>
 </buildCommand>
</buildSpec>
<natures>
 <nature>org.eclipse.cdt.core.cnature</nature>
 <nature>org.eclipse.cdt.core.ccnature</nature>
 <nature>org.eclipse.cdt.managedbuilder.core.managedBuildNature</nature>
 <nature>org.eclipse.cdt.managedbuilder.core.ScannerConfigNature</nature>
</natures>
<linkedResources>"""

    projectFileFooter = """</linkedResources>
    </projectDescription>"""

    with open(os.path.join(destination, ".project"), 'w') as cProjectFile:
        cProjectFile.write(projectFileHeader)
        for name,path in OrderedDict(sorted(dirs, key=lambda t: t[1])).items():
            cProjectFile.write('<link>\n')
            cProjectFile.write(' <name>' + name + '</name>\n')
            cProjectFile.write(' <type>2</type>\n')
            cProjectFile.write(' <location>' + os.path.join(os.getcwd(), path) + '</location>\n')
            cProjectFile.write('</link>\n')
        cProjectFile.write(projectFileFooter)

    if not args.update:
        # Generate buildme
        buildMe = []
        buildMe.append("#!/bin/sh")
        buildMe.append('bob dev "$@" ' + extra + ' ' + quote(project))
        projectCmd = "bob project -n " + extra + " eclipseCdt " + quote(project) + \
            " -u --destination " + quote(destination) + ' --name ' + quote(projectName)
        for i in args.additional_includes:
            projectCmd += " -I " + quote(i)
        buildMe.append(projectCmd)
        generateFile(buildMe, buildMeFile)
        os.chmod(buildMeFile, stat.S_IRWXU | stat.S_IRGRP | stat.S_IWGRP |
            stat.S_IROTH | stat.S_IWOTH)

        # find all executables in package dir
        runTargets = []
        magic = summonMagic()
        if package.getPackageStep().isValid():
            packageDir = package.getPackageStep().getWorkspacePath()
            for root, directory, filenames in os.walk(packageDir):
                for filename in filenames:
                    try:
                        ftype = magic.from_file(os.path.join(root, filename))
                        if 'executable' in ftype and 'x86' in ftype:
                            createLaunchFile(open(os.path.join(destination, filename + ".launch"), 'w'),
                                os.path.join(os.getcwd(), root, filename), projectName)
                    except OSError:
                        pass




# Sample plugin to modify path name calculation
#
# This plugin introduces two new properties to a recipe:
#  - checkoutDir: optional string that is appended to the source directory
#  - platform: optional string that is appenden to the build and dist directories

from os.path import join
from bob.errors import ParseError
from bob.input import PluginState, PluginProperty

def commonFormatter(step, states):
    s = states['pathFmt']
    if step.isCheckoutStep():
        base = step.getPackage().getRecipe().getName()
        ext = s.getCheckoutDir()
        ret = base+"-"+ext if ext else base
    else:
        base = step.getPackage().getName()
        ext = s.getPlatformDir()
        ret = join(base, ext) if ext else base
    return ret.replace('::', "/")

def releaseFormatter(step, states):
    return join("work", commonFormatter(step, states), step.getLabel())

def developFormatter(step, states):
    return join("dev", step.getLabel(), commonFormatter(step, states))

def jenkinsFormatter(step, states):
    return join(commonFormatter(step, states), step.getLabel())

class PathFmtState(PluginState):
    def __init__(self):
        self.__checkoutDir = None
        self.__platformDir = None

    def onEnter(self, env, tools, properties):
        # checkoutDir is always taken from current recipe
        self.__checkoutDir = properties['checkoutDir'].getValue()
        if self.__checkoutDir is not None:
            self.__checkoutDir = env.substitute(self.__checkoutDir, "checkoutDir")

        # platform is passed down to dependencies
        platform = properties['platform']
        if platform.isPresent():
            self.__platformDir = env.substitute(platform.getValue(), "platform")

    def getCheckoutDir(self):
        return self.__checkoutDir

    def getPlatformDir(self):
        return self.__platformDir


class StringProperty(PluginProperty):
    @staticmethod
    def validate(data):
        return isinstance(data, str)


manifest = {
    'apiVersion' : "0.3",
    'hooks' : {
        'releaseNameFormatter' : releaseFormatter,
        'developNameFormatter' : developFormatter,
        'jenkinsNameFormatter' : jenkinsFormatter
    },
    'properties' : {
        "checkoutDir" : StringProperty,
        "platform" : StringProperty
    },
    'state' : {
        "pathFmt" : PathFmtState
    }
}



# Bob build tool
# Copyright (C) 2016-2018  TechniSat Digital GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ...input import RecipeSet
from ...scm import getScm, ScmTaint, ScmStatus
from ...state import BobState
from ...tty import colorize, ERROR, WARNING, EXECUTED, DEFAULT
from ...utils import removePath, processDefines
import argparse
import os

from .builder import LocalBuilder
from .state import DevelopDirOracle

__all__ = ['doClean']

UNKNOWN = ScmStatus(ScmTaint.unknown)

def collectPaths(rootPackage):
    paths = set()
    done = set()

    def walk(package):
        if package._getId() in done: return
        done.add(package._getId())

        checkoutStep = package.getCheckoutStep()
        if checkoutStep.isValid():
            paths.add(checkoutStep.getWorkspacePath())

        # Remove known directories where the digest does not match. The
        # directory state is stored as list where the first entry is the
        # incremental variant id.
        buildStep = package.getBuildStep()
        if buildStep.isValid():
            p = buildStep.getWorkspacePath()
            state = BobState().getDirectoryState(p, False)
            if (state is None) or (buildStep.getVariantId() == state[0]):
                paths.add(p)

        # Remove known directories where the digest does not match.
        packageStep = package.getPackageStep()
        p = packageStep.getWorkspacePath()
        state = BobState().getDirectoryState(p, False)
        if (state is None) or (packageStep.getVariantId() == state):
            paths.add(p)

        for d in package.getDirectDepSteps():
            walk(d.getPackage())

    walk(rootPackage)
    return paths

def checkSCM(workspace, scmDir, scmSpec, verbose):
    if scmDir is None: return True
    if scmSpec is not None:
        status = getScm(scmSpec).status(workspace)
    else:
        status = UNKNOWN

    if verbose:
        flags = str(status)
        if status.error:
            color = ERROR
        elif not status.expendable:
            color = WARNING
        else:
            color = EXECUTED if flags else DEFAULT
        if scmDir != ".":
            workspace = os.path.join(workspace, scmDir)
        print(colorize("STATUS {0: <4} {1}".format(flags, workspace), color))

    return status.expendable

def checkRegularSource(workspace, verbose):
    ret = True
    state = BobState().getDirectoryState(workspace, True)
    for (scmDir, (scmDigest, scmSpec)) in state.items():
        if not checkSCM(workspace, scmDir, scmSpec, verbose):
            ret = False

    return ret

def checkAtticSource(workspace, verbose):
    scmSpec = BobState().getAtticDirectoryState(workspace)
    # We must remove the 'dir' propery if present because the attic directory
    # is already the final directory. Old projects might have scmSpec as None!
    if scmSpec and ('dir' in scmSpec): del scmSpec['dir']
    return checkSCM(workspace, ".", scmSpec, verbose)


def doClean(argv, bobRoot):
    parser = argparse.ArgumentParser(prog="bob clean",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="""Clean unused directories.

This command removes currently unused directories from previous bob dev/build
invocations.  By default only 'build' and 'package' steps are evicted. Adding
'-s' will clean 'checkout' steps too. Make sure that you have checked in (and
pushed) all your changes, tough. When in doubt add '--dry-run' to see what
would get removed without actually deleting that already.
""")

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--develop', action='store_const', const='develop', dest='mode',
        help="Clean developer mode directories (dev/..., default)", default='develop')
    group.add_argument('--release', action='store_const', const='release', dest='mode',
        help="Clean release mode directories (work/...)")
    group.add_argument('--attic', action='store_const', const='attic', dest='mode',
        help="Clean attic directories (dev/.../attic_*)")

    parser.add_argument('-c', dest="configFile", default=[], action='append',
        help="Use config File")
    parser.add_argument('-D', default=[], action='append', dest="defines",
        help="Override default environment variable")
    parser.add_argument('--dry-run', default=False, action='store_true',
        help="Don't delete, just print what would be deleted")
    parser.add_argument('-f', '--force', default=False, action='store_true',
        help="Force deletion of unknown/unclean SCMs")
    parser.add_argument('-s', '--src', default=False, action='store_true',
        help="Clean source workspaces too")

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--sandbox', action='store_true', default=None,
        help="Enable sandboxing")
    group.add_argument('--no-sandbox', action='store_false', dest='sandbox',
        help="Disable sandboxing")

    parser.add_argument('-v', '--verbose', default=False, action='store_true',
        help="Print what is done")
    args = parser.parse_args(argv)

    defines = processDefines(args.defines)

    develop = args.mode != 'release'
    if args.sandbox is None:
        args.sandbox = not develop

    recipes = RecipeSet()
    recipes.defineHook('releaseNameFormatter', LocalBuilder.releaseNameFormatter)
    recipes.defineHook('developNameFormatter', LocalBuilder.developNameFormatter)
    recipes.defineHook('developNamePersister', None)
    recipes.setConfigFiles(args.configFile)
    recipes.parse()

    # Get directory name formatter into shape
    if develop:
        nameFormatter = recipes.getHook('developNameFormatter')
        developPersister = DevelopDirOracle(nameFormatter, recipes.getHook('developNamePersister'))
        nameFormatter = developPersister.getFormatter()
    else:
        # Special read-only "persister" that does create new entries. The
        # actual formatter is irrelevant.
        nameFormatter = LocalBuilder.releaseNameInterrogator
    nameFormatter = LocalBuilder.makeRunnable(nameFormatter)
    packages = recipes.generatePackages(nameFormatter, defines, args.sandbox)
    if develop: developPersister.prime(packages)

    if args.mode == 'attic':
        delPaths = sorted(d for d in BobState().getAtticDirectories()
            if os.path.exists(d) and (args.force or checkAtticSource(d, args.verbose)))
    else:
        if args.mode == 'release':
            # collect all used paths
            usedPaths = collectPaths(packages.getRootPackage())

            # get all known release paths
            allPaths = [ (os.path.join(dir, "workspace"), isSourceDir)
                for (dir, isSourceDir) in BobState().getAllNameDirectores() ]
        elif args.mode == 'develop':
            # collect all used paths
            usedPaths = collectPaths(packages.getRootPackage())

            # Determinte all known develop paths. Bob does not directly store
            # this information.  We start with all known directories and
            # subtract the release paths. Source workspaces are detected by
            # their state being a dict (instead of a bytes object).
            releasePaths = set(os.path.join(dir, "workspace")
                for (dir, isSourceDir) in BobState().getAllNameDirectores())
            allPaths = [
                (dir, isinstance(BobState().getDirectoryState(dir, False), dict))
                for dir in BobState().getDirectories()
                if dir not in releasePaths ]

        # Source workspace policy
        if args.src:
            if args.force:
                mayClean = lambda d: True
            else:
                mayClean = lambda d: checkRegularSource(d, args.verbose)
        else:
            mayClean = lambda d: False

        # Remove non-existent directories and source workspaces that are not
        # allowed to be touched.
        delPaths = sorted(d for (d, isSourceDir) in allPaths
            if (d not in usedPaths) and os.path.exists(d) and
               (not isSourceDir or mayClean(d)))

    # Finally delete unused directories.
    BobState().setAsynchronous()
    try:
        for d in delPaths:
            if args.verbose or args.dry_run:
                print("rm", d)
            if not args.dry_run:
                removePath(d)
                if args.mode == 'attic':
                    BobState().delAtticDirectoryState(d)
                else:
                    BobState().delDirectoryState(d)

        # cleanup BobState() of non-existent directories
        if not args.dry_run:
            for d in BobState().getDirectories():
                if not os.path.exists(d): BobState().delDirectoryState(d)
            for d in BobState().getAtticDirectories():
                if not os.path.exists(d): BobState().delAtticDirectoryState(d)
    finally:
        BobState().setSynchronous()



# Bob build tool
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# Generator for QT-creator project files

import argparse
import sys
import re
import os
import shutil
import stat
import xml.etree.ElementTree
import shutil
from os.path import expanduser
from os.path import join
from bob.utils import removePath, isWindows
from bob.errors import BuildError
from bob.utils import summonMagic, hashFile
from collections import OrderedDict, namedtuple
from bob.tty import colorize
from shlex import quote

# helper to get linux or windows (MSYS2 support) cwd
pwd = None
def cwd():
    global pwd
    if pwd == None and isWindows():
        pwd = os.popen('pwd -W').read().strip()
    return os.getcwd() if not isWindows() else pwd

# scan package recursivelely with its dependencies and build a list of checkout dirs
def getCheckOutDirs(package, excludes, dirs):
    def collect(package, excludes, dirs, steps, processed):
        if package._getId() in processed:
            return
        processed.add(package._getId())
        if package.getCheckoutStep().isValid():
            if package.getCheckoutStep().getVariantId() not in steps:
                steps.add(package.getCheckoutStep().getVariantId())
                dirs.append([package.getName(), package.getCheckoutStep().getWorkspacePath()])

        for d in package.getDirectDepSteps():
            excluded = False
            for e in excludes:
                if (e.search(d.getPackage().getName())):
                    excluded = True
                    break

            if not excluded:
                collect(d.getPackage(), excludes, dirs, steps, processed)

    collect(package, excludes, dirs, set(), set())

def generateFile(entries, fileName):
    try:
        os.remove(fileName)
    except OSError:
        pass
    fileName = open(fileName, "w")
    for e in entries:
        fileName.write(e + "\n")
    fileName.close()

def addRunStep(outFile, num, cmd, directory):
   outFile.write('<valuemap type="QVariantMap" key="ProjectExplorer.Target.RunConfiguration.' + str(num) + '">')
   outFile.write(' <value type="QString" key="ProjectExplorer.CustomExecutableRunConfiguration.Arguments"></value>')
   outFile.write(' <value type="QString" key="ProjectExplorer.CustomExecutableRunConfiguration.Executable">' + cmd + '</value>')
   outFile.write(' <value type="QString" key="ProjectExplorer.CustomExecutableRunConfiguration.WorkingDirectory">' + directory + '</value>')
   outFile.write(' <value type="QString" key="ProjectExplorer.ProjectConfiguration.DefaultDisplayName">' + cmd + '</value>')
   outFile.write(' <value type="QString" key="ProjectExplorer.ProjectConfiguration.DisplayName"></value>')
   outFile.write(' <value type="QString" key="ProjectExplorer.ProjectConfiguration.Id">ProjectExplorer.CustomExecutableRunConfiguration</value>')
   outFile.write(' <value type="bool" key="RunConfiguration.UseCppDebugger">false</value>')
   outFile.write(' <value type="bool" key="RunConfiguration.UseCppDebuggerAuto">true</value>')
   outFile.write(' <value type="bool" key="RunConfiguration.UseMultiProcess">false</value>')
   outFile.write(' <value type="bool" key="RunConfiguration.UseQmlDebugger">false</value>')
   outFile.write(' <value type="bool" key="RunConfiguration.UseQmlDebuggerAuto">true</value>')
   outFile.write('</valuemap>')

RunStep = namedtuple("RunStep", "dir cmd")

def addRunSteps(outFile, runTargets):
    runTargetCnt = 0
    for directory,cmd in runTargets:
        addRunStep(outFile, runTargetCnt, cmd, directory)
        runTargetCnt += 1
    outFile.write('<value type="int" key="ProjectExplorer.Target.RunConfigurationCount">' + str(runTargetCnt) + '</value>')

def addBuildConfig(outFile, num, name, buildArgs, buildMeFile):
    outFile.write('<valuemap type="QVariantMap" key="ProjectExplorer.Target.BuildConfiguration.' + str(num) + '">\n')
    outFile.write('<value type="QString" key="ProjectExplorer.BuildConfiguration.BuildDirectory">' + cwd() + '</value>\n')
    outFile.write('<valuemap type="QVariantMap" key="ProjectExplorer.BuildConfiguration.BuildStepList.0">\n')
    outFile.write(' <valuemap type="QVariantMap" key="ProjectExplorer.BuildStepList.Step.0">\n')
    outFile.write('  <value type="bool" key="ProjectExplorer.BuildStep.Enabled">true</value>\n')
    if isWindows():
        outFile.write('  <value type="QString" key="ProjectExplorer.ProcessStep.Arguments">-msys2 -defterm -no-start -use-full-path -here -c "PATH=' + os.getenv('PATH') + ' sh ' + buildMeFile + ' -v ' + buildArgs + '"</value>\n')
        if os.getenv('WD') is None:
            raise BuildError("Cannot create QtCreator project for Windows! MSYS2 must be started by msys2_shell.cmd script!")
        outFile.write('  <value type="QString" key="ProjectExplorer.ProcessStep.Command">' + os.path.normpath(os.path.join(os.getenv('WD').replace('\\', '/'), '..', '..', 'msys2_shell.cmd')) + '</value>\n')
    else:
        outFile.write('  <value type="QString" key="ProjectExplorer.ProcessStep.Arguments">' + buildArgs + '</value>\n')
        outFile.write('  <value type="QString" key="ProjectExplorer.ProcessStep.Command">' + buildMeFile + '</value>\n')
    outFile.write('  <value type="QString" key="ProjectExplorer.ProcessStep.WorkingDirectory">%{buildDir}</value>\n')
    outFile.write('  <value type="QString" key="ProjectExplorer.ProjectConfiguration.DefaultDisplayName">' + name + '</value>\n')
    outFile.write('  <value type="QString" key="ProjectExplorer.ProjectConfiguration.DisplayName"></value>\n')
    outFile.write('  <value type="QString" key="ProjectExplorer.ProjectConfiguration.Id">ProjectExplorer.ProcessStep</value>\n')
    outFile.write(' </valuemap>\n')
    outFile.write(' <value type="int" key="ProjectExplorer.BuildStepList.StepsCount">1</value>\n')
    outFile.write(' <value type="QString" key="ProjectExplorer.ProjectConfiguration.DefaultDisplayName">Build</value>\n')
    outFile.write(' <value type="QString" key="ProjectExplorer.ProjectConfiguration.DisplayName"></value>\n')
    outFile.write(' <value type="QString" key="ProjectExplorer.ProjectConfiguration.Id">ProjectExplorer.BuildSteps.Build</value>\n')
    outFile.write('</valuemap>\n')
    outFile.write('<value type="int" key="ProjectExplorer.BuildConfiguration.BuildStepListCount">1</value>\n')
    outFile.write('<value type="bool" key="ProjectExplorer.BuildConfiguration.ClearSystemEnvironment">false</value>\n')
    outFile.write('<valuelist type="QVariantList" key="ProjectExplorer.BuildConfiguration.UserEnvironmentChanges">\n')
    outFile.write(' <value type="QString">' + "PATH=" + os.environ['PATH'] + '</value>\n')
    outFile.write('</valuelist>\n')
    outFile.write('<value type="QString" key="ProjectExplorer.ProjectConfiguration.DefaultDisplayName">Vorgabe</value>\n')
    outFile.write('<value type="QString" key="ProjectExplorer.ProjectConfiguration.DisplayName">' + name + '</value>\n')
    outFile.write('<value type="QString" key="ProjectExplorer.ProjectConfiguration.Id">GenericProjectManager.GenericBuildConfiguration</value>\n')
    outFile.write('<valuelist type="QVariantList" key="UserStickyKeys">\n')
    outFile.write(' <value type="QString">ProjectExplorer.BuildConfiguration.UserEnvironmentChanges</value>\n')
    outFile.write(' <value type="QString">UserStickyKeys</value>\n')
    outFile.write('</valuelist>\n')
    outFile.write('</valuemap>\n')

def addBuildSteps(outFile, buildMeFile, buildConfigs):
    addBuildConfig(outFile, 0, "Bob dev ", "" , buildMeFile)
    addBuildConfig(outFile, 1, "Bob dev (force)", "-f", buildMeFile)
    addBuildConfig(outFile, 2, "Bob dev (no checkout)", "-b", buildMeFile)
    addBuildConfig(outFile, 3, "Bob dev (no deps)", "-n", buildMeFile)
    addBuildConfig(outFile, 4, "Bob dev (no checkout, no deps)", "-nb", buildMeFile)
    addBuildConfig(outFile, 5, "Bob dev (no checkout, clean)", "-b --clean", buildMeFile)

    count = 6
    for name,flags in buildConfigs:
        addBuildConfig(outFile, count, name, flags, buildMeFile)
        count += 1

    outFile.write('<value type="int" key="ProjectExplorer.Target.BuildConfigurationCount">' + str(count) + '</value>\n')
    outFile.write('<valuelist type="QVariantList" key="UserStickyKeys">\n')
    outFile.write(' <value type="QString">ProjectExplorer.BuildConfiguration.BuildStepListCount</value>\n')
    outFile.write(' <value type="QString">ProjectExplorer.BuildConfiguration.UserEnvironmentChanges</value>\n')
    outFile.write(' <value type="QString">ProjectExplorer.ProjectConfiguration.DisplayName</value>\n')
    outFile.write('</valuelist>\n')

def compareAndRenameFileIfNotEqual(orig, new):
    if os.path.exists(orig):
        oldHash = hashFile(orig)
        newHash = hashFile(new)
        if (oldHash == newHash):
            os.remove(new)
            return
    os.rename(new, orig)

def parseArgumentLine(line):
    lines = []
    line.lstrip()
    if line.startswith('@'):
        filename = line.lstrip('@')
        try:
            lines = [line.rstrip('\n') for line in open(filename)]
        except IOError:
            raise BuildError("Input file (" + filename + ") could not be read.",
                             help="Specify the argument without leading @, if it is not a file path.")
    else:
        lines.append(line)
    return lines

def qtProjectGenerator(package, argv, extra, bobRoot):
    parser = argparse.ArgumentParser(prog="bob project qt-project", description='Generate QTCreator Project Files')
    parser.add_argument('-u', '--update', default=False, action='store_true',
                        help="Update project files (.files, .includes, .config)")
    parser.add_argument('--buildCfg', action='append', default=[], type=lambda a: a.split("::"),
         help="Adds a new buildconfiguration. Format: <Name>::<flags>")
    parser.add_argument('--overwrite', action='store_true',
        help="Remove destination folder before generating.")
    parser.add_argument('--destination', metavar="DEST",
        help="Destination of project files")
    parser.add_argument('--name', metavar="NAME",
        help="Name of project. Default is complete_path_to_package")
    parser.add_argument('-I', dest="additional_includes", default=[], action='append',
        help="Additional include directories. (added recursive starting from this directory).")
    parser.add_argument('-f', '--filter', metavar="Filter",
        help="File filter. A regex for matching additional files.")
    parser.add_argument('--exclude', default=[], action='append', dest="excludes",
            help="Package filter. A regex for excluding packages in QTCreator.")
    parser.add_argument('--include', default=[], action='append', dest="include",
            help="Include package filter. A regex for including only the specified packages in QTCreator.")
    parser.add_argument('--kit',
        help="Kit to use for this project")
    parser.add_argument('-S', dest="start_includes", default=[], action='append',
        help="Additional include directories, will be placed at the beginning of the include list.")
    parser.add_argument('-C', dest="config_defs", default=[], action='append',
        help="Add line to .config file. Can be used to specify preprocessor defines used by the QTCreator.")

    args = parser.parse_args(argv)
    extra = " ".join(quote(e) for e in extra)

    destination = args.destination
    projectName = args.name

    project = "/".join(package.getStack())

    dirs = []
    excludes = []
    try:
        if args.excludes:
            for e in args.excludes:
                excludes.append(re.compile(e))

        if args.include:
            for e in args.include:
                excludes.append(re.compile(r"^((?!"+e+").)*$"))

        # regex for all source / header files
        source  = re.compile(r"\.[ch](pp)?$")
        include = re.compile(r"\.[h](pp)?$")
        cmake = re.compile(r"\.cmake$")

        if args.filter:
           additionalFiles = re.compile(args.filter)

        # use default kit "Desktop" if no kit is given
        if args.kit is None:
            _kit = re.compile(r".*Desktop.*")
        else:
            _kit = re.compile(r""+args.kit)
    except re.error as e:
        raise ParseError("Invalid regular expression '{}': {}".format(e.pattern), e)

    getCheckOutDirs(package, excludes, dirs)
    if not projectName:
        # use package name for project name
        projectName = package.getName().replace('::', '__')
    if not destination:
        # use package name for project directory
        destination = os.path.join(cwd(), "projects", projectName)
        destination = destination.replace('::', '__')
    if args.overwrite:
        removePath(destination)
    if not os.path.exists(destination):
        os.makedirs(destination)

    # Path structure:
    # destination
    #  -> files
    #  -> project dir
    #     -> symlinks

    # create a 'flat' source tree
    symlinkDir = os.path.join(destination, projectName)
    if not args.update:
       if os.path.exists(symlinkDir):
          shutil.rmtree(symlinkDir)
       os.makedirs(symlinkDir)

    # lists for storing all found sources files / include directories / defines
    sList = []
    hList = []
    sIncList = []
    cList = []

    # now go through all checkout dirs to find all header / include files
    for name,path in OrderedDict(sorted(dirs, key=lambda t: t[1])).items():
        if isWindows():
            name = name.replace('::', '__')
            newPath = os.path.join(cwd(), path)
        else:
            newPath = os.path.join(symlinkDir, name)
            if not args.update: os.symlink(os.path.join(cwd(), path), newPath)
        for root, directories, filenames in os.walk(newPath):
            hasInclude = False
            if (((os.path.sep + '.git' + os.path.sep) in root) or
                ((os.path.sep + '.svn' + os.path.sep) in root)):
                continue
            for filename in filenames:
                if source.search(filename) or cmake.search(filename) or filename == 'CMakeLists.txt':
                    if isWindows():
                        sList.append(os.path.join(root,filename))
                    else:
                        sList.append(os.path.join(cwd(), os.path.join(root,filename)))
                if args.filter and additionalFiles.search(filename):
                    if isWindows():
                        sList.append(os.path.join(root,filename))
                    else:
                        sList.append(os.path.join(cwd(), os.path.join(root,filename)))
                if not hasInclude and include.search(filename):
                    hasInclude = True
            if hasInclude:
                oldPath = ""
                # need to recursively add all directories from the include up to cwd to get includes like
                # <a/b/c.h> resolved even if there is no include in 'a'
                relativePath = root[len(newPath):]
                for path in relativePath.split(os.path.sep):
                    includePath = os.path.join(newPath, oldPath, path)
                    if not includePath in hList:
                        hList.append(includePath)
                    oldPath = os.path.join(oldPath, path)

    for i in args.additional_includes:
        for e in parseArgumentLine(i):
            if os.path.exists(e):
                for root, directories, filenames in os.walk(e):
                    hList.append(os.path.join(e,root))

    # compose start includes
    for i in args.start_includes:
        for e in parseArgumentLine(i):
            if os.path.exists(i):
                sIncList.append(e)

    # compose content of .config file (preprocessor defines)
    for i in args.config_defs:
        for e in parseArgumentLine(i):
             cList.append(e)

    id = None
    name = None
    kits = []
    try:
        if isWindows():
            profiles = xml.etree.ElementTree.parse(os.path.join(os.getenv('APPDATA'), "QtProject/qtcreator/profiles.xml")).getroot()
        else:
            profiles = xml.etree.ElementTree.parse(os.path.join(expanduser('~'), ".config/QtProject/qtcreator/profiles.xml")).getroot()
        for profile in profiles:
            for valuemap in profile:
                id = None
                name = None
                for value in valuemap.findall('value'):
                    if (value.attrib.get('key') == 'PE.Profile.Id'):
                        id = str(value.text)
                    if (value.attrib.get('key') == 'PE.Profile.Name'):
                        name = str(value.text)
                    if (id is not None) and (name is not None) and (_kit.search(name)):
                        kits.append([name, id])
                        break
    except FileNotFoundError:
        # if kits are generared using sdk tool they are stored somewhere else... (/usr/share...)
        pass

    if (len(kits) == 0):
        if (args.kit is None):
            raise BuildError("No kit found!",
                help = "Run again with '--kit' and specify a kit or generate a Desktop kit, which is used by default.")
        kitName = args.kit
        kitId = args.kit
    else:
        if (len(kits) > 1):
            print(colorize("Warning: {} kits found. Using '{}'.".format(len(kits), str(kits[0][0])), "33"))
        kitName = kits[0][0]
        kitId = kits[0][1]

    buildMeFile = os.path.join(destination, "buildme")

    # Generate the includes file. Create a temporary file first and compare it with the old one.
    # only use the new one if different to prevent reindexing after each build.
    includesFile = os.path.join(destination, projectName + ".includes")
    generateFile(sIncList + sorted(hList, key=lambda file: (os.path.dirname(file))), includesFile + ".new")
    compareAndRenameFileIfNotEqual(includesFile, includesFile + ".new")

    # Generate files file
    filesFile = os.path.join(destination, projectName  + ".files")
    sList.append(buildMeFile)
    generateFile(sorted(sList, key=lambda file: (os.path.dirname(file), os.path.basename(file))), filesFile + ".new")
    compareAndRenameFileIfNotEqual(filesFile, filesFile + ".new")

    # Generate a .config file
    configFile = os.path.join(destination, projectName  + ".config")
    generateFile(cList, configFile + ".new")
    compareAndRenameFileIfNotEqual(configFile, configFile + ".new")

    if not args.update:
        # Generate Buildme.sh
        buildMe = []
        buildMe.append("#!/bin/sh")
        buildMe.append('bob dev "$@" ' + extra + ' ' + quote(project))
        projectCmd = "bob project -n " + extra + " qt-creator " + quote(project) + \
            " -u --destination " + quote(destination) + ' --name ' + quote(projectName)
        # only add arguments which are relevant for .files or .includes. All other files are only modified if not build with
        # update only.
        for i in args.additional_includes:
            projectCmd += " -I " + quote(i)
        if args.filter:
            projectCmd += " --filter " + quote(args.filter)
        if args.excludes:
            for e in args.excludes:
                projectCmd += " --exclude " + quote(e)
        if args.include:
            for e in args.include:
                projectCmd += "--include " + quote(e)
        for e in args.start_includes:
            projectCmd += " -S " + quote(e)
        for e in args.config_defs:
            projectCmd += " -C " + quote(e)

        buildMe.append(projectCmd)
        generateFile(buildMe, buildMeFile)
        os.chmod(buildMeFile, stat.S_IRWXU | stat.S_IRGRP | stat.S_IWGRP |
            stat.S_IROTH | stat.S_IWOTH)
        # Generate creator file
        creatorFile = os.path.join(destination, projectName  + ".creator")
        generateFile([], creatorFile)
        # Generate the creator.shared file using a template and modify some settings
        sharedFile = os.path.join(destination, projectName  + ".creator.shared")

        template_head = """<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE QtCreatorProject>
<qtcreator>
 <data>
  <variable>ProjectExplorer.Project.ActiveTarget</variable>
  <value type="int">0</value>
 </data>
 <data>
  <variable>ProjectExplorer.Project.EditorSettings</variable>
  <valuemap type="QVariantMap">
   <value type="bool" key="EditorConfiguration.AutoIndent">true</value>
   <value type="bool" key="EditorConfiguration.AutoSpacesForTabs">false</value>
   <value type="bool" key="EditorConfiguration.CamelCaseNavigation">true</value>
   <valuemap type="QVariantMap" key="EditorConfiguration.CodeStyle.0">
    <value type="QString" key="language">Cpp</value>
    <valuemap type="QVariantMap" key="value">
     <value type="QByteArray" key="CurrentPreferences">CppGlobal</value>
    </valuemap>
   </valuemap>
   <valuemap type="QVariantMap" key="EditorConfiguration.CodeStyle.1">
    <value type="QString" key="language">QmlJS</value>
    <valuemap type="QVariantMap" key="value">
     <value type="QByteArray" key="CurrentPreferences">QmlJSGlobal</value>
    </valuemap>
   </valuemap>
   <value type="int" key="EditorConfiguration.CodeStyle.Count">2</value>
   <value type="QByteArray" key="EditorConfiguration.Codec">UTF-8</value>
   <value type="bool" key="EditorConfiguration.ConstrainTooltips">false</value>
   <value type="int" key="EditorConfiguration.IndentSize">4</value>
   <value type="bool" key="EditorConfiguration.KeyboardTooltips">false</value>
   <value type="int" key="EditorConfiguration.MarginColumn">80</value>
   <value type="bool" key="EditorConfiguration.MouseHiding">true</value>
   <value type="bool" key="EditorConfiguration.MouseNavigation">true</value>
   <value type="int" key="EditorConfiguration.PaddingMode">1</value>
   <value type="bool" key="EditorConfiguration.ScrollWheelZooming">true</value>
   <value type="bool" key="EditorConfiguration.ShowMargin">false</value>
   <value type="int" key="EditorConfiguration.SmartBackspaceBehavior">0</value>
   <value type="bool" key="EditorConfiguration.SmartSelectionChanging">true</value>
   <value type="bool" key="EditorConfiguration.SpacesForTabs">true</value>
   <value type="int" key="EditorConfiguration.TabKeyBehavior">0</value>
   <value type="int" key="EditorConfiguration.TabSize">4</value>
   <value type="bool" key="EditorConfiguration.UseGlobal">true</value>
   <value type="int" key="EditorConfiguration.Utf8BomBehavior">1</value>
   <value type="bool" key="EditorConfiguration.addFinalNewLine">true</value>
   <value type="bool" key="EditorConfiguration.cleanIndentation">true</value>
   <value type="bool" key="EditorConfiguration.cleanWhitespace">true</value>
   <value type="bool" key="EditorConfiguration.inEntireDocument">false</value>
  </valuemap>
 </data>
 <data>
  <variable>ProjectExplorer.Project.PluginSettings</variable>
  <valuemap type="QVariantMap"/>
 </data>"""

        template_foot = """
 <data>
  <variable>ProjectExplorer.Project.TargetCount</variable>
  <value type="int">1</value>
 </data>
 <data>
  <variable>ProjectExplorer.Project.Updater.FileVersion</variable>
  <value type="int">18</value>
 </data>
 <data>
  <variable>Version</variable>
  <value type="int">18</value>
 </data>
</qtcreator>"""

        # find all executables in package dir
        runTargets = []
        magic = summonMagic()
        if package.getPackageStep().isValid():
            packageDir = package.getPackageStep().getWorkspacePath()
            for root, directory, filenames in os.walk(packageDir):
                for filename in filenames:
                    try:
                        ftype = magic.from_file(os.path.join(root, filename))
                        if 'executable' in ftype and 'x86' in ftype:
                            runTargets.append(RunStep(os.path.join(cwd(), root), filename))
                    except OSError:
                        pass

        with open(sharedFile, 'w') as sharedFile:
            sharedFile.write(template_head)
            sharedFile.write(' <data>\n')
            sharedFile.write('  <variable>ProjectExplorer.Project.Target.0</variable>\n')
            sharedFile.write('  <valuemap type="QVariantMap">\n')
            sharedFile.write('   <value type="QString" key="ProjectExplorer.ProjectConfiguration.DefaultDisplayName">' + kitName + '</value>\n')
            sharedFile.write('   <value type="QString" key="ProjectExplorer.ProjectConfiguration.DisplayName">' + kitName + '</value>\n')
            sharedFile.write('   <value type="QString" key="ProjectExplorer.ProjectConfiguration.Id">' + kitId + '</value>\n')
            sharedFile.write('   <value type="int" key="ProjectExplorer.Target.ActiveBuildConfiguration">0</value>\n')
            sharedFile.write('   <value type="int" key="ProjectExplorer.Target.ActiveDeployConfiguration">0</value>\n')
            sharedFile.write('   <value type="int" key="ProjectExplorer.Target.ActiveRunConfiguration">0</value>\n')
            addBuildSteps(sharedFile, buildMeFile, args.buildCfg)
            sharedFile.write('   <value type="int" key="ProjectExplorer.Target.DeployConfigurationCount">0</value>\n')
            sharedFile.write('   <valuemap type="QVariantMap" key="ProjectExplorer.Target.PluginSettings"/>\n')
            addRunSteps(sharedFile, runTargets)
            sharedFile.write('  </valuemap>\n')
            sharedFile.write(' </data>\n')
            sharedFile.write(template_foot)





# Bob build tool
# Copyright (C) 2017  Jan Klötzke
#
# SPDX-License-Identifier: GPL-3.0-or-later

from . import BOB_INPUT_HASH
from .errors import BuildError, ParseError
from .scm import GitAudit, SvnAudit, UrlAudit, ImportAudit, auditFromData
from .utils import asHexStr, hashFile, binStat
from datetime import datetime, timezone
import gzip
import hashlib
import io
import json
import platform
import pickle
import schema
import struct

def digestMap(m, h):
    h.update(struct.pack("<BI", 1, len(m)))
    for (k,v) in sorted(m.items()):
        digestString(k, h)
        digestData(v, h)

def digestString(s, h):
    h.update(struct.pack("<BI", 2, len(s)))
    h.update(s.encode('utf8'))

def digestData(d, h):
    if isinstance(d, str):
        digestString(d, h)
    elif isinstance(d, dict):
        digestMap(d, h)
    elif isinstance(d, list):
        h.update(struct.pack("<BI", 3, len(d)))
        for i in d: digestData(i, h)
    elif isinstance(d, int):
        h.update(struct.pack("<Bq", 4, d))
    elif isinstance(d, bool):
        h.update(struct.pack("<B?", 5, d))
    elif isinstance(d, bytes):
        h.update(struct.pack("<BI", 6, len(d)))
        h.update(d)
    else:
        assert False, "Cannot digest " + str(type(d))

class HexValidator:
    def validate(self, data):
        try:
            return bytes.fromhex(data)
        except ValueError:
            raise schema.SchemaUnexpectedTypeError("not valid hex str", None)

class Artifact:

    SCMS = {
        'git' : GitAudit,
        'svn' : SvnAudit,
        'url' : UrlAudit,
        'import' : ImportAudit,
    }

    SCHEMA = schema.Schema({
        "variant-id" : HexValidator(),
        "build-id" : HexValidator(),
        "artifact-id" : HexValidator(),
        "result-hash" : HexValidator(),
        "meta" : { schema.Optional(str) : str },
        "build" : {
            'sysname'  : str,
            'nodename' : str,
            'release'  : str,
            'version'  : str,
            'machine'  : str,
            'date'     : str,
            schema.Optional('os-release') : str,
        },
        "env" : str,
        schema.Optional('metaEnv') : { schema.Optional(str) : str },
        "scms" : [ dict ],
        schema.Optional("recipes") : dict,
        "dependencies" : {
            schema.Optional('args') : [ HexValidator() ],
            schema.Optional('tools') : { str : HexValidator() },
            schema.Optional('sandbox') : HexValidator()
        }
    })

    def __init__(self):
        self.reset(b'\x00' * 20, b'\x00' * 20, b'\x00' * 20)

    @classmethod
    def fromData(cls, data):
        artifact = cls()
        artifact.load(data)
        return artifact

    @classmethod
    def __getOsRelease(cls):
        try:
            ret = cls.__osRelease
        except AttributeError:
            try:
                with open("/etc/os-release") as f:
                    ret = f.read()
            except OSError:
                ret = None
            cls.__osRelease = ret

        return ret

    def __calculate(self):
        if self.__id is not None: return
        d = self.__dump()
        h = hashlib.sha1()
        digestData(d, h)
        self.__id = h.digest()

    def reset(self, variantId, buildId, resultHash):
        self.__variantId = variantId
        self.__buildId = buildId
        self.__resultHash = resultHash
        self.__recipes = None
        self.__defines = {}
        u = platform.uname()
        self.__build = {
            'sysname'  : u.system,
            'nodename' : u.node,
            'release'  : u.release,
            'version'  : u.version,
            'machine'  : u.machine,
            'date'     : datetime.now(timezone.utc).isoformat(),
        }
        osRelease = self.__getOsRelease()
        if osRelease is not None:
            self.__build['os-release'] = osRelease
        self.__env = ""
        self.__metaEnv = {}
        self.__scms = []
        self.__deps = []
        self.__tools = {}
        self.__sandbox = None
        self.__id = None

    def load(self, data):
        self.__id = None
        self.__variantId = data["variant-id"]
        self.__buildId = data["build-id"]
        self.__resultHash = data["result-hash"]

        recipes = data.get("recipes")
        if recipes is not None:
            self.__recipes = auditFromData(recipes)
        else:
            self.__recipes = None

        self.__defines = data["meta"]
        self.__build = data["build"]
        self.__env = data["env"]
        self.__metaEnv = data.get("metaEnv", {})

        self.__scms = []
        scms = data["scms"]
        for i in scms:
            self.__scms.append(auditFromData(i))

        deps = data["dependencies"]
        self.__deps = deps.get("args", [])
        self.__tools = deps.get("tools", {})
        self.__sandbox = deps.get("sandbox")

        # validate id
        self.__calculate()
        if self.__id != data["artifact-id"]:
            raise ParseError("Corrupt Audit! Artifact-Id does not match!")

    def dump(self):
        d = self.__dump()
        d['artifact-id'] = asHexStr(self.getId())
        return d

    def __dump(self):
        dependencies = {}
        if self.__deps:
            dependencies["args"] = [ asHexStr(i) for i in self.__deps ]
        if self.__tools:
            dependencies["tools"] = { n : asHexStr(t) for (n,t) in self.__tools.items() }
        if self.__sandbox:
            dependencies["sandbox"] = asHexStr(self.__sandbox)

        ret = {
            "variant-id" : asHexStr(self.__variantId),
            "build-id" : asHexStr(self.__buildId),
            "result-hash" : asHexStr(self.__resultHash),
            "meta" : self.__defines,
            "build" : self.__build,
            "env" : self.__env,
            "scms" : [ s.dump() for s in self.__scms ],
            "dependencies" : dependencies
        }

        if self.__metaEnv:
            ret[ "metaEnv"] = self.__metaEnv

        if self.__recipes is not None:
            ret["recipes"] = self.__recipes.dump()

        return ret

    def setRecipes(self, recipes):
        self.__recipes = recipes

    def setEnv(self, env):
        try:
            with open(env) as f:
                self.__env = f.read()
        except OSError as e:
            raise ParseError("Error reading environment: " + str(e))
        self.__id = None

    def addDefine(self, name, value):
        self.__defines[name] = value
        self.__id = None

    async def addScm(self, name, workspace, dir, extra):
        scm = Artifact.SCMS.get(name)
        if scm is None:
            raise BuildError("Cannot handle SCM: " + name)
        self.__scms.append(await scm.fromDir(workspace, dir, extra))
        self.__id = None

    def addTool(self, name, tool):
        self.__tools[name] = tool
        self.__id = None

    def addMetaEnv(self, var, value):
        self.__metaEnv[var] = value

    def setSandbox(self, sandbox):
        self.__sandbox = sandbox
        self.__id = None

    def addArg(self, arg):
        self.__deps.append(arg)
        self.__id = None

    def getId(self):
        self.__calculate()
        return self.__id

    def getBuildId(self):
        return self.__buildId

    def getReferences(self):
        ret = set()
        for i in self.__deps: ret.add(i)
        if self.__sandbox: ret.add(self.__sandbox)
        for i in self.__tools.values(): ret.add(i)
        return ret

    def getMetaData(self):
        return self.__defines

    def getBuildInfo(self):
        return self.__build

    def getMetaEnv(self):
        return self.__metaEnv

class Audit:
    SCHEMA = schema.Schema({
        'artifact' : Artifact.SCHEMA,
        'references' : [ Artifact.SCHEMA ]
    })

    def __init__(self):
        self.__artifact = Artifact()
        self.__references = {}

    @classmethod
    def fromFile(cls, file):
        try:
            cacheName = file + ".pickle"
            cacheKey = binStat(file) + BOB_INPUT_HASH
            with open(cacheName, "rb") as f:
                persistedCacheKey = f.read(len(cacheKey))
                if cacheKey == persistedCacheKey:
                    return pickle.load(f)
        except (EOFError, OSError, pickle.UnpicklingError) as e:
            pass

        audit = cls()
        try:
            with gzip.open(file, 'rb') as gzf:
                audit.load(gzf, file)
            with open(cacheName, "wb") as f:
                f.write(cacheKey)
                pickle.dump(audit, f, -1)
        except OSError as e:
            raise ParseError("Error loading audit: " + str(e))
        return audit

    @classmethod
    def fromByteStream(cls, stream, name):
        audit = cls()
        audit.load(stream, name)
        return audit

    @classmethod
    def create(cls, variantId, buildId, resultHash):
        audit = cls()
        audit.reset(variantId, buildId, resultHash)
        return audit

    def __merge(self, other):
        self.__references.update(other.__references)
        self.__references[other.getId()] = other.__artifact

    def __validate(self):
        done = set()
        refs = self.__artifact.getReferences()
        while refs:
            curId = refs.pop()
            cur = self.__references.get(curId)
            if cur is None:
                raise ParseError("Incomplete audit: missing " + asHexStr(curId))
            for dep in cur.getReferences():
                if dep not in done: refs.add(dep)
            done.add(curId)

    def load(self, file, name):
        try:
            tree = json.load(io.TextIOWrapper(file, encoding='utf8'))
            tree = Audit.SCHEMA.validate(tree)
            self.__artifact = Artifact.fromData(tree["artifact"])
            self.__references = {
                r["artifact-id"] : Artifact.fromData(r) for r in tree["references"]
            }
        except schema.SchemaError as e:
            raise ParseError(name + ": Invalid audit record: " + str(e))
        except ValueError as e:
            raise ParseError(name + ": Invalid json: " + str(e))
        self.__validate()

    def save(self, file):
        tree = {
            "artifact" : self.__artifact.dump(),
            "references" : [ a.dump() for a in self.__references.values() ]
        }
        try:
            with gzip.open(file, 'wb', 6) as gzf:
                json.dump(tree, io.TextIOWrapper(gzf, encoding='utf8'))

            cacheName = file + ".pickle"
            cacheKey = binStat(file) + BOB_INPUT_HASH
            with open(cacheName, "wb") as f:
                f.write(cacheKey)
                pickle.dump(self, f, -1)

        except OSError as e:
            raise BuildError("Cannot write audit: " + str(e))

    def reset(self, variantId, buildId, resultHash):
        self.__artifact.reset(variantId, buildId, resultHash)
        self.__references = {}

    def getId(self):
        return self.__artifact.getId()

    def getArtifact(self, aid=None):
        if aid:
            return self.__references[aid]
        else:
            return self.__artifact

    def getReferencedBuildIds(self):
        ret = set()
        refs = self.__artifact.getReferences()
        while refs:
            artifact = self.__references[refs.pop()]
            if artifact.getMetaData()["step"] == "dist":
                ret.add(artifact.getBuildId())
            else:
                refs.update(artifact.getReferences())
        return sorted(ret)

    def setRecipesAudit(self, recipes):
        self.__artifact.setRecipes(recipes)

    def setRecipesData(self, xml):
        self.__artifact.setRecipes(auditFromData(xml))

    def setEnv(self, env):
        self.__artifact.setEnv(env)

    def addDefine(self, name, value):
        self.__artifact.addDefine(name, value)

    async def addScm(self, name, workspace, dir, extra):
        await self.__artifact.addScm(name, workspace, dir, extra)

    def addTool(self, name, tool):
        audit = Audit.fromFile(tool)
        self.__merge(audit)
        self.__artifact.addTool(name, audit.getId())

    def addMetaEnv(self, var, value):
        self.__artifact.addMetaEnv(var, value)

    def setSandbox(self, sandbox):
        audit = Audit.fromFile(sandbox)
        self.__merge(audit)
        self.__artifact.setSandbox(audit.getId())

    def addArg(self, arg):
        audit = Audit.fromFile(arg)
        self.__merge(audit)
        self.__artifact.addArg(audit.getId())




# Bob build tool
# Copyright (C) 2016-2020 The BobBuildTool Contributors
#
# SPDX-License-Identifier: GPL-3.0-or-later

from .. import BOB_VERSION
from ..errors import BuildError, ParseError
from ..stringparser import IfExpression
from ..utils import asHexStr, hashFile, isWindows
from .scm import Scm, ScmAudit
import asyncio
import concurrent
import contextlib
import hashlib
import os, os.path
import re
import schema
import shutil
import signal
import ssl
import stat
import tempfile
import time
import urllib.error
import urllib.parse
import urllib.request


WEEKDAYNAME = ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun']
MONTHNAME = ['Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep',
    'Oct', 'Nov', 'Dec']

def isYounger(update, existing):
    if not os.path.exists(existing): return True
    return os.stat(update).st_mtime > os.stat(existing).st_mtime

def time2HTTPDate(timestamp):
    """Return the timestamp formatted as RFC7231 Section 7.1.1.1 conformant string."""
    year, month, day, hh, mm, ss, wd, y, z = time.gmtime(timestamp)
    return "%s, %02d %3s %4d %02d:%02d:%02d GMT" % (WEEKDAYNAME[wd], day,
        MONTHNAME[month-1], year, hh, mm, ss)

def startsWithDrive(url):
    if len(url) < 2: return False
    if url[1] != ':': return False
    return ((url[0] >= 'A') and (url[0] <= 'Z')) or ((url[0] >= 'a') and (url[0] <= 'z'))

def parseUrl(url):
    r"""Parse URL and apply Windows quirks.

    On Windows we allow the user to specify the following paths as URLs. They
    are no valid URLs but we still accept them. cURL will work on them too
    (plus many others)...

     * C:\foo.bar
     * C:/foo.bar
     * file:///C:/foo.bar
     * file:///C:\foo.bar
     * \\server\path
     * file:///\\server\path

    We explicitly do not allow relative or absolute paths. Only fully qualified
    paths are accepted. See
    https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file for
    more information. The following paths are rejected:

     * C:tmp.txt
     * tmp.txt
     * \tmp.txt
     * /tmp.txt
     * file:///tmp.txt
     * file:///C:tmp.txt
     * file:///\tmp.txt
    """

    # Only require some special processing on Windows. Unix URLs just work...
    if not isWindows():
        return urllib.parse.urlparse(url)

    # Does it start with a drive letter like "C:…"?
    if startsWithDrive(url):
        # Convert slashes to backslashes and make sure it's a fully qualified
        # path.
        path = url.replace('/', '\\')
        if not path[2:].startswith("\\"):
            raise ValueError("URL must be a fully qualified path name")
        return urllib.parse.ParseResult('file', '', path, '', '', '')
    elif url.startswith("\\\\"):
        # No slash conversion on UNC paths!
        return urllib.parse.ParseResult('file', '', url, '', '', '')

    # Regular parsing. Bail out if it is not a local path.
    url = urllib.parse.urlparse(url)
    if url.scheme == '':
        raise ValueError("Neither a fully qualified path nor a URL scheme given")
    if url.scheme != 'file':
        return url

    # The host part is checked in the generic code. We just take care of the
    # path here.
    if not url.path.startswith("/"):
        raise ValueError("Invalid path")

    path = url.path[1:]
    if path.startswith("\\\\"):
        pass
    elif startsWithDrive(path) and (path[2:].startswith("\\") or path[2:].startswith("/")):
        path = path.replace('/', '\\')
    else:
        raise ValueError("URL must be a fully qualified path name")

    # looks legit
    return urllib.parse.ParseResult(url.scheme, url.netloc,path, '', '', '')


class UrlScm(Scm):

    SCHEMA = schema.Schema({
        'scm' : 'url',
        'url' : str,
        schema.Optional('dir') : str,
        schema.Optional('if') : schema.Or(str, IfExpression),
        schema.Optional('digestSHA1') : str,
        schema.Optional('digestSHA256') : str,
        schema.Optional('extract') : schema.Or(bool, str),
        schema.Optional('fileName') : str,
        schema.Optional('stripComponents') : int,
        schema.Optional('sslVerify') : bool,
    })

    EXTENSIONS = [
        (".tar.gz",    "tar"),
        (".tar.xz",    "tar"),
        (".tar.bz2",   "tar"),
        (".tar.bzip2", "tar"),
        (".tgz",       "tar"),
        (".tar",       "tar"),
        (".gz",        "gzip"),
        (".xz",        "xz"),
        (".7z",        "7z"),
        (".zip",       "zip"),
    ]

    EXTRACTORS = {
        "tar"  : [
            ("tar", ["-x", "--no-same-owner", "--no-same-permissions", "-f", "{}"], "--strip-components={}"),
            ("python", ["-m", "tarfile", "-e", "{}"], None),
        ],
        "gzip" : [
            ("gunzip", ["-kf", "{}"], None),
        ],
        "xz" : [
            ("unxz", ["-kf", "{}"], None),
        ],
        "7z" : [
            ("7z", ["x", "-y", "{}"], None),
        ],
        "zip" : [
            ("unzip", ["-o", "{}"], None),
            ("python", ["-m", "zipfile", "-e", "{}", "."], None),
        ],
    }

    def __init__(self, spec, overrides=[], tidy=None):
        super().__init__(spec, overrides)
        self.__url = spec["url"]
        self.__digestSha1 = spec.get("digestSHA1")
        if self.__digestSha1:
            # validate digest
            if re.match("^[0-9a-f]{40}$", self.__digestSha1) is None:
                raise ParseError("Invalid SHA1 digest: " + str(self.__digestSha1))
        self.__digestSha256 = spec.get("digestSHA256")
        if self.__digestSha256:
            # validate digest
            if re.match("^[0-9a-f]{64}$", self.__digestSha256) is None:
                raise ParseError("Invalid SHA256 digest: " + str(self.__digestSha256))
        self.__dir = spec.get("dir", ".")
        self.__fn = spec.get("fileName")
        if not self.__fn:
            url = self.__url
            if isWindows():
                # On Windows we're allowed to provide native paths with
                # backslashes.
                url = url.replace('\\', '/')
            self.__fn = url.split("/")[-1]
        self.__extract = spec.get("extract", "auto")
        self.__tidy = tidy
        self.__strip = spec.get("stripComponents", 0)
        self.__sslVerify = spec.get('sslVerify', True)

    def getProperties(self):
        ret = super().getProperties()
        ret.update({
            'scm' : 'url',
            'url' : self.__url,
            'digestSHA1' : self.__digestSha1,
            'digestSHA256' : self.__digestSha256,
            'dir' : self.__dir,
            'fileName' : self.__fn,
            'extract' : self.__extract,
            'stripComponents' : self.__strip,
            'sslVerify' : self.__sslVerify,
        })
        return ret

    def _download(self, destination):
        # restore signals to default so that Ctrl+C kills us
        signal.signal(signal.SIGINT, signal.SIG_DFL)

        headers = {}
        headers["User-Agent"] = "BobBuildTool/{}".format(BOB_VERSION)
        context = None if self.__sslVerify else ssl.SSLContext(ssl.PROTOCOL_SSLv23)
        if os.path.isfile(destination) and self.__url.startswith("http"):
            # Try to avoid download if possible
            headers["If-Modified-Since"] = time2HTTPDate(os.stat(destination).st_mtime)

        tmpFileName = None
        req = urllib.request.Request(url=self.__url, headers=headers)
        try:
            with contextlib.closing(urllib.request.urlopen(req, context=context)) as rsp:
                with tempfile.NamedTemporaryFile(dir=os.path.dirname(destination), delete=False) as f:
                    tmpFileName = f.name
                    read = 0
                    while True:
                        buf = rsp.read(16384)
                        if not buf:
                            break
                        read += len(buf)
                        f.write(buf)

                if "content-length" in rsp.info():
                    expected = int(rsp.info()["Content-Length"])
                    if expected > read:
                        return "Response too short: {} < {} (bytes)".format(read, expected)

                # Atomically move file to destination. Set explicit mode to
                # retain Bob 0.15 behaviour.
                os.chmod(tmpFileName, stat.S_IREAD|stat.S_IWRITE)
                os.replace(tmpFileName, destination)
                tmpFileName = None

        except urllib.error.HTTPError as e:
            if e.code != 304:
                return "HTTP error {}: {}".format(e.code, e.reason)
        finally:
            if tmpFileName is not None:
                os.remove(tmpFileName)

        return None

    async def invoke(self, invoker):
        os.makedirs(invoker.joinPath(self.__dir), exist_ok=True)
        workspaceFile = os.path.join(self.__dir, self.__fn)
        destination = invoker.joinPath(self.__dir, self.__fn)

        # Download only if necessary
        if not self.isDeterministic() or not os.path.isfile(destination):
            try:
                url = parseUrl(self.__url)
            except ValueError as e:
                invoker.fail(str(e))

            if url.scheme in ['', 'file']:
                # Verify that host name is empty or "localhost"
                if url.netloc not in ['', 'localhost']:
                    invoker.fail("Bad/unsupported URL: invalid host name: " + url.netloc)
                # Local files: copy only if newer (u), target never is a directory (T)
                if isYounger(url.path, destination):
                    if os.path.isdir(destination):
                        invoker.fail("Destination", destination, "is an existing directory!")
                    invoker.trace("<cp>", url.path, workspaceFile)
                    shutil.copy(url.path, destination)
            elif url.scheme in ["http", "https", "ftp"]:
                invoker.trace("<wget>", self.__url, ">", workspaceFile)
                loop = asyncio.get_event_loop()
                try:
                    err = await loop.run_in_executor(None, UrlScm._download, self, destination)
                    if err:
                        invoker.fail(err)
                except (concurrent.futures.CancelledError, concurrent.futures.process.BrokenProcessPool):
                    invoker.fail("Download interrupted!")
            else:
                invoker.fail("Unsupported URL scheme: " + url.scheme)


        # Always verify file hashes
        if self.__digestSha1:
            invoker.trace("<sha1sum>", workspaceFile)
            d = hashFile(destination, hashlib.sha1).hex()
            if d != self.__digestSha1:
                invoker.fail("SHA1 digest did not match! expected:", self.__digestSha1, "got:", d)
        if self.__digestSha256:
            invoker.trace("<sha256sum>", workspaceFile)
            d = hashFile(destination, hashlib.sha256).hex()
            if d != self.__digestSha256:
                invoker.fail("SHA256 digest did not match! expected:", self.__digestSha256, "got:", d)

        # Run optional extractors
        extractors = self.__getExtractors()
        canary = invoker.joinPath(self.__dir, "." + self.__fn + ".extracted")
        if extractors and isYounger(destination, canary):
            for cmd in extractors:
                if shutil.which(cmd[0]) is None: continue
                await invoker.checkCommand(cmd, cwd=self.__dir)
                invoker.trace("<touch>", canary)
                with open(canary, "wb") as f:
                    pass
                os.utime(canary)
                break
            else:
                executor.fail("No suitable extractor found!")

    def asDigestScript(self):
        """Return forward compatible stable string describing this url.

        The format is "digest dir extract" if a SHA checksum was specified.
        Otherwise it is "url dir extract". A "s#" is appended if leading paths
        are stripped where # is the number of stripped elements.
        """
        return ( self.__digestSha256 if self.__digestSha256
                 else (self.__digestSha1 if self.__digestSha1 else self.__url)
                    ) + " " + os.path.join(self.__dir, self.__fn) + " " + str(self.__extract) + \
                    ( " s{}".format(self.__strip) if self.__strip > 0 else "" )

    def getDirectory(self):
        return self.__dir if self.__tidy else os.path.join(self.__dir, self.__fn)

    def isDeterministic(self):
        return (self.__digestSha1 is not None) or (self.__digestSha256 is not None)

    def getAuditSpec(self):
        return ("url", os.path.join(self.__dir, self.__fn),
                {"url" : self.__url})

    def hasLiveBuildId(self):
        return self.isDeterministic()

    async def predictLiveBuildId(self, step):
        return self.calcLiveBuildId(None)

    def calcLiveBuildId(self, workspacePath):
        if self.__digestSha256:
            return bytes.fromhex(self.__digestSha256)
        elif self.__digestSha1:
            return bytes.fromhex(self.__digestSha1)
        else:
            return None

    def getLiveBuildIdSpec(self, workspacePath):
        if self.__digestSha256:
            return "=" + self.__digestSha256
        elif self.__digestSha1:
            return "=" + self.__digestSha1
        else:
            return None

    def __getExtractors(self):
        extractors = None
        if self.__extract in ["yes", "auto", True]:
            for (ext, tool) in UrlScm.EXTENSIONS:
                if self.__fn.endswith(ext):
                    extractors = UrlScm.EXTRACTORS[tool]
                    break
            if not extractors and self.__extract != "auto":
                raise ParseError("Don't know how to extract '"+self.__fn+"' automatically.")
        elif self.__extract in UrlScm.EXTRACTORS:
            extractors = UrlScm.EXTRACTORS[tool]
        elif self.__extract not in ["no", False]:
            raise ParseError("Invalid extract mode: " + self.__extract)

        if extractors is None:
            return []

        ret = []
        for extractor in extractors:
            if self.__strip > 0:
                if extractor[2] is None:
                    continue
                strip = [extractor[2].format(self.__strip)]
            else:
                strip = []
            ret.append([extractor[0]] + [a.format(self.__fn) for a in extractor[1]] + strip)

        if not ret:
            raise ParseError("Extractor does not support 'stripComponents'!")

        return ret


class UrlAudit(ScmAudit):

    SCHEMA = schema.Schema({
        'type' : 'url',
        'dir' : str,
        'digest' : {
            'algorithm' : 'sha1',
            'value' : str
        },
        schema.Optional('url') : str, # Added in Bob 0.16
    })

    async def _scanDir(self, workspace, dir, extra):
        self.__dir = dir
        self.__hash = asHexStr(hashFile(os.path.join(workspace, dir)))
        self.__url = extra.get("url")

    def _load(self, data):
        self.__dir = data["dir"]
        self.__hash = data["digest"]["value"]
        self.__url = data.get("url")

    def dump(self):
        ret = {
            "type" : "url",
            "dir" : self.__dir,
            "digest" : {
                "algorithm" : "sha1",
                "value" : self.__hash
            }
        }
        if self.__url is not None:
            ret["url"] = self.__url

        return ret



# Bob build tool
# Copyright (C) 2016-2018  TechniSat Digital GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ...errors import BobError
import os
import sqlite3


class DevelopDirOracle:
    """
    Calculate directory names for develop mode.

    If an external "persister" is used we just cache the calculated values. We
    don't know it's behaviour and have to re-calculate everything from scratch
    to be on the safe side.

    The internal algorithm creates a separate directory for every recipe and
    step variant. Only identical steps of the same recipe are put into the
    same directory. In contrast to the releaseNamePersister() identical steps
    of different recipes are put into distinct directories. If the recipes are
    changed we keep existing mappings that still match the base directory.

    Populating the database is done by traversing all packages and invoking the
    name formatter for the visited packages. In case of the external persister
    the result is directly cached. For the internal algorithm it has to be done
    in two passes. The first pass collects all packages and their base
    directories, possibly re-using old matches if possible. The second pass
    assigns the final directory names to all other entries and writes them into
    the database.
    """

    def __init__(self, formatter, externalPersister):
        self.__formatter = formatter
        self.__externalPersister = externalPersister(formatter) \
            if externalPersister is not None else None
        self.__dirs = {}
        self.__known = {}
        self.__visited = set()
        self.__ready = False

    def __fmt(self, step, props):
        key = step.getPackage().getRecipe().getName().encode("utf8") + step.getVariantId()

        # Always look into the database first. We almost always need it.
        self.__db.execute("SELECT dir FROM dirs WHERE key=?", (key,))
        path = self.__db.fetchone()
        path = path[0] if path is not None else None

        # If we're ready we just interrogate the database.
        if self.__ready:
            assert path is not None, "{} missing".format(key)
            return path

        # Make sure to process each key only once. A key might map to several
        # directories. We have to make sure to take only the first one, though.
        if key in self.__visited: return
        self.__visited.add(key)

        # If an external persister is used we just call it and save the result.
        if self.__externalPersister is not None:
            self.__known[key] = self.__externalPersister(step, props)
            return

        # Try to find directory in database. If we find some the prefix has to
        # match. Otherwise schedule for number assignment in next round by
        # __writeBack(). The final path is then not decided yet.
        baseDir = self.__formatter(step, props)
        if (path is not None) and path.startswith(baseDir):
            self.__known[key] = path
        else:
            self.__dirs.setdefault(baseDir, []).append(key)

    def __touch(self, package, done):
        """Run through all dependencies and invoke name formatter.

        Traversal is done on package level to gather all reachable packages of
        the query language.
        """
        key = package._getId()
        if key in done: return
        done.add(key)

        # Traverse direct package dependencies only to keep the recipe order.
        # Because we start from the root we are guaranteed to see all packages.
        for d in package.getDirectDepSteps():
            self.__touch(d.getPackage(), done)

        # Calculate the paths of all steps
        package.getPackageStep().getWorkspacePath()
        package.getBuildStep().getWorkspacePath()
        package.getCheckoutStep().getWorkspacePath()

    def __writeBack(self):
        """Write calculated directories into database.

        We have to write known entries and calculate new sub-directory numbers
        for new entries.
        """
        # clear all mappings
        self.__db.execute("DELETE FROM dirs")

        # write kept entries
        self.__db.executemany("INSERT INTO dirs VALUES (?, ?)", self.__known.items())
        knownDirs = set(self.__known.values())

        # Add trailing number to new entries. Make sure they don't collide with
        # kept entries...
        for baseDir,keys in self.__dirs.items():
            num = 1
            for key in keys:
                while True:
                    path = os.path.join(baseDir, str(num))
                    num += 1
                    if path in knownDirs: continue
                    self.__db.execute("INSERT INTO dirs VALUES (?, ?)", (key, path))
                    break

        # Clear intermediate variables to save memory.
        self.__dirs = {}
        self.__known = {}
        self.__visited = set()

    def __openAndRefresh(self, cacheKey, rootPackage):
        self.__db = db = sqlite3.connect(".bob-dev-dirs.sqlite3", isolation_level=None).cursor()
        db.execute("CREATE TABLE IF NOT EXISTS meta(key PRIMARY KEY, value)")
        db.execute("CREATE TABLE IF NOT EXISTS dirs(key PRIMARY KEY, dir)")

        # Check if recipes were changed.
        db.execute("BEGIN")
        db.execute("SELECT value FROM meta WHERE key='vsn'")
        vsn = db.fetchone()
        if (vsn is None) or (vsn[0] != cacheKey):
            self.__touch(rootPackage, set())
            self.__writeBack()
            db.execute("INSERT OR REPLACE INTO meta VALUES ('vsn', ?)", (cacheKey,))
            # Commit and start new read-only transaction
            db.execute("END")
            db.execute("BEGIN")

    def prime(self, packages):
        try:
            self.__openAndRefresh(packages.getCacheKey(),
                packages.getRootPackage())
        except sqlite3.Error as e:
            raise BobError("Cannot save directory mapping: " + str(e))
        self.__ready = True

    def getFormatter(self):
        return self.__fmt




# Bob build tool
# Copyright (C) 2016-2019  TechniSat Digital GmbH
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ...input import RecipeSet
from ...scm import getScm, ScmTaint, ScmStatus
from ...state import BobState
from ...tty import colorize, ERROR, WARNING, EXECUTED, DEFAULT, SKIPPED, \
    IMPORTANT, NORMAL, INFO, DEBUG, TRACE, HEADLINE
from ...utils import joinLines, processDefines
from textwrap import indent
import argparse
import os

from .builder import LocalBuilder
from .state import DevelopDirOracle

__all__ = ['doStatus']

# Flag to headline verbosity. The description is shown on the next level.
FLAG_TO_VERBOSITY = {
    ScmTaint.attic          : NORMAL,
    ScmTaint.collides       : NORMAL,       # not modified, but will break the build
    ScmTaint.error          : IMPORTANT,    # error, modified
    ScmTaint.modified       : NORMAL,       # modified
    ScmTaint.new            : NORMAL,
    ScmTaint.overridden     : DEBUG,
    ScmTaint.switched       : NORMAL,       # modified
    ScmTaint.unknown        : NORMAL,       # cannot tell, could be modified
    ScmTaint.unpushed_main  : NORMAL,       # modified
    ScmTaint.unpushed_local : INFO,         # not modified but user may loose data
}
assert set(FLAG_TO_VERBOSITY.keys()) == set(ScmTaint)

class PackagePrinter:
    def __init__(self, verbose, showClean, showOverrides, checkoutStep = None):
        self.verbose = verbose
        self.flagVerbosity = FLAG_TO_VERBOSITY.copy()
        self.showClean = showClean
        if showOverrides: self.flagVerbosity[ScmTaint.overridden] = NORMAL
        self.headerShown = checkoutStep is None
        self.checkoutStep = checkoutStep

    def __printHeader(self):
        if not self.headerShown:
            print(">>", colorize("/".join(self.checkoutStep.getPackage().getStack()),
                                 HEADLINE))
            self.headerShown = True

    def __printStatus(self, flags, message, color):
        print(colorize("   STATUS {0: <4} {1}".format(flags, message), color))

    def show(self, status, dir):
        detailedFlags = { flag for flag,severity in self.flagVerbosity.items()
            if severity < self.verbose }

        # Determine severity of headline. If showClean start directly at NORMAL
        # level.
        severity = NORMAL if self.showClean else DEBUG
        for flag in status.flags:
            severity = min(self.flagVerbosity[flag], severity)

        flags = str(status)
        if status.error:
            color = ERROR
        elif status.dirty or (status.flags & {ScmTaint.unknown, ScmTaint.collides}):
            color = WARNING
        elif flags:
            color = EXECUTED
        else:
            color = DEFAULT

        if severity <= self.verbose:
            self.__printHeader()
            self.__printStatus(flags, dir, color)
            description = status.description(detailedFlags)
            if description:
                for line in description.splitlines():
                    print('   ' + line)

    def skipped(self):
        # skipped workspaces are shown only on '-vvv' at least
        if TRACE <= self.verbose:
            self.__printHeader()
            self.__printStatus("",
                "skipped ({} does not exist)".format(self.checkoutStep.getWorkspacePath()),
                SKIPPED)

ATTIC = ScmStatus(ScmTaint.attic,
    description="> Recipe changed. Will be moved to attic on next checkout.")
UNKNOWN = ScmStatus(ScmTaint.unknown,
    description="> Workspace too old. Cannot determine status.")

class Printer:
    def __init__(self, recurse, verbose, showClean, showOverrides, showAttic):
        self.recurse = recurse
        self.verbose = verbose
        self.showClean = showClean
        self.showOverrides = showOverrides
        self.doneSteps = set()
        self.donePackages = set()
        self.showAttic = showAttic

    def __showCheckoutStep(self, pp, checkoutStep):
        workspace = checkoutStep.getWorkspacePath()
        oldCheckoutState = BobState().getDirectoryState(workspace, True)
        checkoutState = checkoutStep.getScmDirectories()
        scms = { scm.getDirectory() : scm for scm in checkoutStep.getScmList() }
        result = {}

        # First scan old checkout state. This is what the user is most
        # interested in. The recipe might have changed compared to the
        # persisted state!
        for (scmDir, (scmDigest, scmSpec)) in oldCheckoutState.items():
            if scmDir is None: continue # checkoutScript state -> ignored
            if not os.path.exists(os.path.join(workspace, scmDir)): continue

            if scmDigest == checkoutState.get(scmDir, (None, None))[0]:
                # The digest still matches -> use recipe values
                status = scms[scmDir].status(workspace)
            elif scmSpec is not None:
                # New project that kept scm spec -> compare with that and mark
                # as attic
                status = getScm(scmSpec).status(workspace)
                status.merge(ATTIC)
            else:
                # Don't know anything about it except that this will be moved
                # to the attic
                status = ScmStatus()
                status.merge(ATTIC)
                status.merge(UNKNOWN)

            result[scmDir] = status

        # Additionally scan current checkout state to find new checkouts and
        # determinte override status.
        for scmDir in checkoutState.keys():
            status = result.setdefault(scmDir, ScmStatus(ScmTaint.new))
            if (ScmTaint.new in status.flags) and os.path.exists(os.path.join(workspace, scmDir)):
                status.add(ScmTaint.collides,
                    "> Collides with existing file in workspace.")
            elif ScmTaint.attic in status.flags:
                status.add(ScmTaint.new)

            # The override status is taken from the recipe scm. This is
            # independent of any actual checkout.
            overrides = scms[scmDir].getActiveOverrides()
            for o in overrides:
                status.add(ScmTaint.overridden, joinLines("> Overridden by:",
                    indent(str(o), '   ')))

        for (scmDir, status) in sorted(result.items()):
            pp.show(status, os.path.join(workspace, scmDir))

    def __showAtticDirs(self, pp, prefix=""):
        for d in sorted(BobState().getAtticDirectories()):
            if not os.path.isdir(d):
                BobState().delAtticDirectoryState(d)
                continue
            if not d.startswith(prefix): continue

            scmSpec = BobState().getAtticDirectoryState(d)
            if scmSpec is not None:
                # We must remove the 'dir' propery if present because the attic
                # directory is already the final directory.
                if 'dir' in scmSpec: del scmSpec['dir']
                status = getScm(scmSpec).status(d)
            else:
                status = UNKNOWN

            pp.show(status, d)

    def showPackage(self, package):
        if package._getId() in self.donePackages: return
        self.donePackages.add(package._getId())

        checkoutStep = package.getCheckoutStep()
        if checkoutStep.isValid() and (checkoutStep.getVariantId() not in self.doneSteps):
            pp = PackagePrinter(self.verbose, self.showClean, self.showOverrides,
                checkoutStep)
            workspace = checkoutStep.getWorkspacePath()
            if workspace is not None:
                if os.path.isdir(workspace):
                    self.__showCheckoutStep(pp, checkoutStep)
                else:
                    pp.skipped()
                if self.showAttic:
                    # The last path element (/workspace) must be removed because
                    # attics are located next to the workspace, not inside it.
                    self.__showAtticDirs(pp, os.path.dirname(workspace))
        self.doneSteps.add(checkoutStep.getVariantId())

        if self.recurse:
            for d in package.getDirectDepSteps():
                self.showPackage(d.getPackage())

    def showAllDirs(self, showAttic):
        pp = PackagePrinter(self.verbose, self.showClean, self.showOverrides)
        for workspace in sorted(BobState().getDirectories()):
            dirState = BobState().getDirectoryState(workspace, False)
            # Only the checkout state is stored as dict. Use that to find out
            # which are the right directories.
            if not isinstance(dirState, dict):
                continue

            if not os.path.isdir(workspace):
                BobState().delDirectoryState(workspace)
                continue

            # Upgrade from old format without scmSpec. Drop None dir.
            dirState = sorted(
                (dir, state) if isinstance(state, tuple) else (dir, (state, None))
                for dir,state in dirState.items() if dir is not None)
            for (scmDir, (scmDigest, scmSpec)) in dirState:
                scmDir = os.path.join(workspace, scmDir)
                if scmSpec is not None:
                    status = getScm(scmSpec).status(workspace)
                else:
                    status = UNKNOWN
                pp.show(status, scmDir)

        if showAttic:
            self.__showAtticDirs(pp)


def doStatus(argv, bobRoot):
    parser = argparse.ArgumentParser(prog="bob status", description='Show SCM status')
    parser.add_argument('packages', nargs='*', help="(Sub-)packages")

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--develop', action='store_true',  dest='develop', help="Use developer mode", default=True)
    group.add_argument('--release', action='store_false', dest='develop', help="Use release mode")

    parser.add_argument('-c', dest="configFile", default=[], action='append',
        help="Use config File")
    parser.add_argument('-D', default=[], action='append', dest="defines",
        help="Override default environment variable")

    parser.add_argument('--attic', action='store_true',
        help="Additionally look in/for attic directories")
    parser.add_argument('-r', '--recursive', default=False, action='store_true',
                        help="Recursively display dependencies")

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--sandbox', action='store_true', help="Enable sandboxing")
    group.add_argument('--no-sandbox', action='store_false', dest='sandbox', help="Disable sandboxing")
    parser.set_defaults(sandbox=None)

    parser.add_argument('--show-clean', action='store_true',
        help="Show SCM status even if checkout is unmodified")
    parser.add_argument('--show-overrides', action='store_true',
        help="Show SCM status if affected by an scmOverrides")
    parser.add_argument('-v', '--verbose', default=NORMAL, action='count',
        help="Increase verbosity (may be specified multiple times)")
    args = parser.parse_args(argv)

    if args.sandbox == None:
        args.sandbox = not args.develop

    defines = processDefines(args.defines)

    recipes = RecipeSet()
    recipes.defineHook('releaseNameFormatter', LocalBuilder.releaseNameFormatter)
    recipes.defineHook('developNameFormatter', LocalBuilder.developNameFormatter)
    recipes.defineHook('developNamePersister', None)
    recipes.setConfigFiles(args.configFile)
    recipes.parse()

    if args.develop:
        # Develop names are stable. All we need to do is to replicate build's algorithm,
        # and when we produce a name, check whether it exists.
        nameFormatter = recipes.getHook('developNameFormatter')
        developPersister = DevelopDirOracle(nameFormatter, recipes.getHook('developNamePersister'))
        nameFormatter = developPersister.getFormatter()
    else:
        # Release names are taken from persistence.
        nameFormatter = LocalBuilder.releaseNameInterrogator
    nameFormatter = LocalBuilder.makeRunnable(nameFormatter)

    packages = recipes.generatePackages(nameFormatter, defines, args.sandbox)
    if args.develop: developPersister.prime(packages)

    # Dummy query of attic directories. Will warn if project directory was
    # created before Bob 0.15 where they were not tracked!
    if args.attic:
        BobState().getAtticDirectories()

    # Set BobState into asynchronous mode because we might remove many entries
    # if their directories do not exist anymore.
    BobState().setAsynchronous()
    try:
        printer = Printer(args.recursive, args.verbose, args.show_clean,
            args.show_overrides, args.attic)
        if args.packages:
            for p in args.packages:
                for package in packages.queryPackagePath(p):
                    printer.showPackage(package)
        else:
            printer.showAllDirs(args.attic)
    finally:
        BobState().setSynchronous()




# Bob build tool
# Copyright (C) 2016  Jan Klötzke
#
# SPDX-License-Identifier: GPL-3.0-or-later

from ..errors import BuildError
from ..stringparser import IfExpression
from ..utils import joinLines, check_output
from .scm import Scm, ScmAudit, ScmTaint, ScmStatus
from shlex import quote
from textwrap import indent
import os, os.path
import schema
import subprocess
from xml.etree import ElementTree

class SvnScm(Scm):

    SCHEMA = schema.Schema({
        'scm' : 'svn',
        'url' : str,
        schema.Optional('dir') : str,
        schema.Optional('if') : schema.Or(str, IfExpression),
        schema.Optional('revision') : schema.Or(int, str),
        schema.Optional('sslVerify') : bool,
    })

    def __init__(self, spec, overrides=[]):
        super().__init__(spec, overrides)
        self.__url = spec["url"]
        self.__dir = spec.get("dir", ".")
        self.__revision = spec.get("revision")
        self.__sslVerify = spec.get('sslVerify', True)

    def getProperties(self):
        ret = super().getProperties()
        ret.update({
            'scm' : 'svn',
            "url" : self.__url,
            "dir" : self.__dir,
            'sslVerify' : self.__sslVerify,
        })
        if self.__revision:
            ret["revision"] = self.__revision
        return ret

    async def invoke(self, invoker):
        options = [ "--non-interactive" ]
        if not self.__sslVerify:
            options += [ "--trust-server-cert-failures=unknown-ca,cn-mismatch,expired,not-yet-valid,other" ]
        if self.__revision:
            options += [ "-r", str(self.__revision) ]

        if os.path.isdir(invoker.joinPath(self.__dir, ".svn")):
            if "/tags/" not in self.__url:
                await invoker.checkCommand(["svn", "up"] + options, cwd=self.__dir)
        else:
            await invoker.checkCommand(["svn", "co"] + options + [self.__url, self.__dir])

    def asDigestScript(self):
        """Return forward compatible stable string describing this svn module.

        The module is represented as "url[@rev] > dir".
        """
        return (self.__url + ( ("@"+str(self.__revision)) if self.__revision else "" ) + " > "
                + self.__dir)

    def asJenkins(self, workPath, credentials, options):
        scm = ElementTree.Element("scm", attrib={
            "class" : "hudson.scm.SubversionSCM",
            "plugin" : "subversion@2.4.5",
        })

        locations = ElementTree.SubElement(scm, "locations")
        location = ElementTree.SubElement(locations,
            "hudson.scm.SubversionSCM_-ModuleLocation")

        url = self.__url
        if self.__revision:
            url += ( "@" + str(self.__revision) )

        ElementTree.SubElement(location, "remote").text = url
        credentialsId = ElementTree.SubElement(location, "credentialsId")
        if credentials: credentialsId.text = credentials
        ElementTree.SubElement(location, "local").text = (
            os.path.normpath(os.path.join(workPath, self.__dir)) )
        ElementTree.SubElement(location, "depthOption").text = "infinity"
        ElementTree.SubElement(location, "ignoreExternalsOption").text = "true"

        ElementTree.SubElement(scm, "excludedRegions")
        ElementTree.SubElement(scm, "includedRegions")
        ElementTree.SubElement(scm, "excludedUsers")
        ElementTree.SubElement(scm, "excludedRevprop")
        ElementTree.SubElement(scm, "excludedCommitMessages")
        ElementTree.SubElement(scm, "workspaceUpdater",
            attrib={"class":"hudson.scm.subversion.UpdateUpdater"})
        ElementTree.SubElement(scm, "ignoreDirPropChanges").text = "false"
        ElementTree.SubElement(scm, "filterChangelog").text = "false"

        return scm

    def getDirectory(self):
        return self.__dir

    def isDeterministic(self):
        return str(self.__revision).isnumeric()

    def hasJenkinsPlugin(self):
        return True

    def callSubversion(self, workspacePath, *args):
        cmdLine = ['svn']
        cmdLine.extend(args)
        cwd = os.path.join(workspacePath, self.__dir)
        try:
            output = subprocess.check_output(cmdLine, cwd=cwd,
                universal_newlines=True, stderr=subprocess.DEVNULL)
        except subprocess.CalledProcessError as e:
            raise BuildError("svn error:\n Directory: '{}'\n Command: '{}'\n'{}'".format(
                cwd, " ".join(cmdLine), e.output.rstrip()))
        except OSError as e:
            raise BuildError("Error calling svn: " + str(e))
        return output.strip()

    def status(self, workspacePath):
        status = ScmStatus()
        try:
            output = self.callSubversion(workspacePath, 'status')
            if output:
                status.add(ScmTaint.modified, joinLines("> modified:", indent(output, '   ')))

            output = self.callSubversion(workspacePath, 'info', '--xml')
            info = ElementTree.fromstring(output)
            entry = info.find('entry')
            url = entry.find('url').text
            revision = entry.attrib['revision']

            if self.__url != url:
                status.add(ScmTaint.switched,
                    "> URL: configured: '{}', actual: '{}'".format(self.__url, url))
            if self.__revision is not None and int(revision) != int(self.__revision):
                status.add(ScmTaint.switched,
                    "> revision: configured: {}, actual: {}".format(self.__revision, revision))

        except BuildError as e:
            status.add(ScmTaint.error, e.slogan)

        return status

    def getAuditSpec(self):
        return ("svn", self.__dir, {})


class SvnAudit(ScmAudit):

    SCHEMA = schema.Schema({
        'type' : 'svn',
        'dir' : str,
        'url' : str,
        'revision' : int,
        'dirty' : bool,
        'repository' : {
            'root' : str,
            'uuid' : str
        }
    })

    async def _scanDir(self, workspace, dir, extra):
        self.__dir = dir
        try:
            info = ElementTree.fromstring(await check_output(
                ["svn", "info", "--xml", dir],
                cwd=workspace, universal_newlines=True))
            self.__url = info.find('entry/url').text
            self.__revision = int(info.find('entry').get('revision'))
            self.__repoRoot = info.find('entry/repository/root').text
            self.__repoUuid = info.find('entry/repository/uuid').text

            status = await check_output(["svn", "status", dir],
                cwd=workspace, universal_newlines=True)
            self.__dirty = status != ""
        except subprocess.CalledProcessError as e:
            raise BuildError("Svn audit failed: " + str(e))
        except OSError as e:
            raise BuildError("Error calling svn: " + str(e))
        except ElementTree.ParseError as e:
            raise BuildError("Invalid XML received from svn")

    def _load(self, data):
        self.__dir = data["dir"]
        self.__url = data["url"]
        self.__revision = data["revision"]
        self.__dirty = data["dirty"]
        self.__repoRoot = data["repository"]["root"]
        self.__repoUuid = data["repository"]["uuid"]

    def dump(self):
        return {
            "type" : "svn",
            "dir" : self.__dir,
            "url" : self.__url,
            "revision" : self.__revision,
            "dirty" :  self.__dirty,
            "repository" : {
                "root" :  self.__repoRoot,
                "uuid" :  self.__repoUuid,
            }
        }

    def getStatusLine(self):
        return self.__url + "@" + str(self.__revision) + ("-dirty" if self.__dirty else "")



# Bob build tool
# Copyright (C) 2020  Jan Klötzke
#
# SPDX-License-Identifier: GPL-3.0-or-later

from . import BOB_INPUT_HASH
from .errors import ParseError
from .utils import escapePwsh, quotePwsh, isWindows, asHexStr
from .utils import joinScripts, sliceString
from base64 import b64encode
from enum import Enum
from glob import glob
from shlex import quote
from textwrap import dedent
import hashlib
import json
import os
import sys

__all__ = ['ScriptLanguage', 'getLanguage', 'StepSpec']

BASH_FINGERPRINT_SNIPPETS = [
    ("bob-libc-version", r"""
bob-libc-version()
{
    if ! type -p ${1:-${CC:-cc}} >/dev/null ; then
        echo "No C-Compiler!" >&2
        return 1
    fi

    # Machine type is important (e.g. x86_64)
    uname -m

    # Try glibc first
    cat >conftest.c <<EOF
#include <stdio.h>
#include <gnu/libc-version.h>
int main(){ printf("glibc %s\n", gnu_get_libc_version()); return 0; }
EOF
    if ${1:-${CC:-cc}} -o conftest conftest.c >/dev/null ; then
        ./conftest && return 0
    fi

    # Maybe musl libc? Link a simple program and extract runtime loader. On
    # musl the runtime loader is executable and outputs its version.
    cat >conftest.c <<EOF
int main(){ return 0; }
EOF
    if ! ${1:-${CC:-cc}} -o conftest conftest.c >/dev/null ; then
        echo "The C-Compiler does not seem to work... :(" >&2
        return 1
    fi

    DL=$(readelf -p .interp ./conftest | sed -n -e '/ld-musl/s/[^/]*\(\/.*\)/\1/p')
    if [[ -x $DL ]] ; then
        $DL 2>&1 || true
        return 0
    fi

    # Uhh?
    echo "Unsupported system. Please consider submitting your OS configuration for inclusion." >&2
    return 1
}
"""),

    ("bob-libstdc++-version", r"""
bob-libstdc++-version()
{
    if ! type -p ${1:-${CXX:-c++}} >/dev/null ; then
        echo "No C++-Compiler!" >&2
        return 1
    fi

    # Machine type is important (e.g. x86_64)
    uname -m

    cat >conftest.cpp <<EOF
#include <iostream>
int main(int /*argc*/, char ** /*argv*/)
{
    int ret = 1;
#ifdef __GLIBCXX__
    std::cout << "libstdc++ " << __GLIBCXX__ << &std::endl;
    ret = 0;
#endif
#ifdef _LIBCPP_VERSION
    std::cout << "libc++ " << _LIBCPP_VERSION << &std::endl;
    ret = 0;
#endif
    return ret;
}
EOF
    ${1:-${CXX:-c++}} -o conftest conftest.cpp >/dev/null
    ./conftest
}
"""),

    ("bob-hash-libraries", r"""
bob-hash-libraries()
{
    declare -a opts=( -o canary -xc - )
    local i

    for i in "$@" ; do
        opts+=( -l "$i" )
    done

    echo "int main(){return 0;}" | ${CC:-cc} "${opts[@]}"
    sha1sum $(ldd canary | grep -o -e '/[^[:space:]]*' | sort -u)
}
"""),
]


class ScriptLanguage(Enum):
    BASH = 'bash'
    PWSH = 'PowerShell'


class IncludeResolver:
    def __init__(self, fileLoader, baseDir, origText, sourceName, varBase):
        self.fileLoader = fileLoader
        self.baseDir = baseDir
        self.sourceName = sourceName
        self.varBase = varBase
        self.__incDigests = [ hashlib.sha1(origText.encode('utf8')).digest().hex() ]

    def __getitem__(self, item):
        mode = item[0]
        item = item[1:]
        content = []
        try:
            paths = sorted(glob(os.path.join(self.baseDir, item)))
            if not paths:
                raise ParseError("No files matched in include pattern '{}'!"
                    .format(item))
            for path in paths:
                content.append(self.fileLoader(path))
        except OSError as e:
            raise ParseError("Error including '"+item+"': " + str(e))
        content = b''.join(content)

        self.__incDigests.append(hashlib.sha1(content).digest().hex())
        if mode == '<':
            ret = self._includeFile(content)
        else:
            assert mode == "'"
            ret = self._includeLiteral(content)

        return ret

    def _includeFile(self, content):
        raise NotImplementedError()

    def _includeLiteral(self, content):
        raise NotImplementedError()

    def _resolveContent(self, result):
        raise NotImplementedError()

    def resolve(self, result):
        return (self._resolveContent(result), "\n".join(self.__incDigests))

# For each supported language the following runtime environments must be
# considered:
#
#  * Native POSIX platform
#  * Native Windows platform
#  * MSYS2 POSIX platform running on Windows
#
# When passing paths and/or writing them into scripts the paths may have to be
# converted.

class BashResolver(IncludeResolver):
    def __init__(self, fileLoader, baseDir, origText, sourceName, varBase):
        super().__init__(fileLoader, baseDir, origText, sourceName, varBase)
        self.prolog = []
        self.count = 0

    def _includeFile(self, content):
        var = "_{}{}".format(self.varBase, self.count)
        self.count += 1
        self.prolog.extend([
            "{VAR}=$(mktemp)".format(VAR=var),
            "_BOB_TMP_CLEANUP+=( ${VAR} )".format(VAR=var),
            "base64 -d > ${VAR} <<EOF".format(VAR=var)])
        self.prolog.extend(sliceString(b64encode(content).decode("ascii"), 76))
        self.prolog.append("EOF")
        return "${" + var + "}"

    def _includeLiteral(self, content):
        return quote(content.decode('utf8'))

    def _resolveContent(self, result):
        tail = ["_BOB_SOURCES[$LINENO]=" + quote(self.sourceName), result]
        return "\n".join(self.prolog + tail)


class BashLanguage:
    index = ScriptLanguage.BASH
    glue = "\ncd \"${BOB_CWD}\"\n"
    Resolver = BashResolver

    # When we try to execute bash scripts on Windows we have to take into
    # account that MSYS2 uses unix paths internally. So if we have
    #
    #   C:\foo\bar
    #
    # we have to transform this into
    #
    #  /c/foo/bar
    #
    if sys.platform == "win32":
        @staticmethod
        def __munge(p):
            if p[1:3] == ":\\":
                return "/" + p[0].lower() + "/" + p[3:].replace('\\', '/')
            else:
                return p.replace('\\', '/')
    else:
        @staticmethod
        def __munge(p):
            return p

    @staticmethod
    def __formatSetup(spec, keepEnv):
        env = { key: quote(value) for (key, value) in spec.env.items() }
        env.update({
            "PATH": ":".join(
                [quote(BashLanguage.__munge(os.path.abspath(p))) for p in spec.paths] +
                (["$PATH"] if not spec.hasSandbox else
                 [quote(p) for p in spec.sandboxPaths])
            ),
            "LD_LIBRARY_PATH": ":".join(
                quote(BashLanguage.__munge(os.path.abspath(p))) for p in spec.libraryPaths
            ),
            "BOB_CWD": quote(BashLanguage.__munge(os.path.abspath(spec.workspaceExecPath))),
        })

        ret = [
            "# Automatically generated file!",
            "# It's content will be overwritten every time the step is run.",
            "",
        ]

        if keepEnv:
            # Parse global config files if env should be kept
            ret.append(dedent("""\
                [[ -e /etc/bash.bashrc ]] && source /etc/bash.bashrc
                [[ -e ~/.bashrc ]] && source ~/.bashrc
                """))

        ret.extend([
            "# Special Bob array variables:",
            "declare -A BOB_ALL_PATHS=( {} )".format(" ".join(sorted(
                [ "[{}]={}".format(quote(name), quote(BashLanguage.__munge(os.path.abspath(path))))
                    for name,path in spec.allPaths ] ))),
            "declare -A BOB_DEP_PATHS=( {} )".format(" ".join(sorted(
                [ "[{}]={}".format(quote(name), quote(BashLanguage.__munge(os.path.abspath(path))))
                    for name,path in spec.depPaths ] ))),
            "declare -A BOB_TOOL_PATHS=( {} )".format(" ".join(sorted(
                [ "[{}]={}".format(quote(name), quote(BashLanguage.__munge(os.path.abspath(path))))
                    for name,path in spec.toolPaths ] ))),
            "",
            "# Environment:",
            "\n".join("export {}={}".format(k, v) for (k,v) in sorted(env.items()))
        ])
        return "\n".join(ret)

    @staticmethod
    def __formatScript(spec):
        colorize = not spec.isJenkins
        envFile = "/bob/env" if spec.hasSandbox else os.path.abspath(spec.envFile)
        ret = [
            BashLanguage.__formatSetup(spec, False),
            "",
            dedent("""\
                # Setup
                declare -p > {ENV_FILE}
                cd \"${{BOB_CWD}}\"
                """.format(ENV_FILE=quote(BashLanguage.__munge(envFile)))),
            dedent("""\
                # Error handling
                bob_handle_error()
                {
                    set +x"""),
            '    echo "\x1b[31;1mStep failed with return status $1; Command:\x1b[0;31m ${BASH_COMMAND}\x1b[0m"' if colorize \
                else '    echo "Step failed with return status $1; Command: ${BASH_COMMAND}"',
            dedent("""\
                    echo "Call stack (most recent call first)"
                    i=0
                    while caller $i >/dev/null ; do
                            j=${BASH_LINENO[$i]}
                            while [[ $j -ge 0 && -z ${_BOB_SOURCES[$j]:+true} ]] ; do
                                    : $(( j-- ))
                            done
                            echo "    #$i: ${_BOB_SOURCES[$j]}, line $(( BASH_LINENO[$i] - j )), in ${FUNCNAME[$((i+1))]}"
                            : $(( i++ ))
                    done

                    exit $1
                }
                declare -A _BOB_SOURCES=( [0]="Bob prolog" )
                trap 'bob_handle_error $? >&2 ; exit 99' ERR
                trap 'for i in "${_BOB_TMP_CLEANUP[@]-}" ; do /bin/rm -f "$i" ; done' EXIT
                set -o errtrace -o nounset -o pipefail
                """),
            "# Recipe script",
            spec.script,
        ]
        return "\n".join(ret)

    @staticmethod
    def __scriptFilePaths(spec, tmpDir):
        if spec.hasSandbox:
            execScriptFile = "/.script"
            realScriptFile = spec.scriptHint or os.path.join(tmpDir, ".script")
        else:
            execScriptFile = spec.scriptHint or os.path.join(tmpDir, "script")
            realScriptFile = execScriptFile
        return (os.path.abspath(realScriptFile), os.path.abspath(execScriptFile))

    @staticmethod
    def setupShell(spec, tmpDir, keepEnv):
        realScriptFile, execScriptFile = BashLanguage.__scriptFilePaths(spec, tmpDir)
        with open(realScriptFile, "w") as f:
            f.write(BashLanguage.__formatSetup(spec, keepEnv))

        args = ["bash", "--rcfile", BashLanguage.__munge(execScriptFile), "-s", "--"]
        args.extend(BashLanguage.__munge(os.path.abspath(a)) for a in spec.args)
        return (realScriptFile, execScriptFile, args)

    @staticmethod
    def setupCall(spec, tmpDir, keepEnv, trace):
        realScriptFile, execScriptFile = BashLanguage.__scriptFilePaths(spec, tmpDir)
        with open(realScriptFile, "w") as f:
            f.write(BashLanguage.__formatScript(spec))

        args = ["bash"]
        if trace: args.append("-x")
        args.extend(["--", BashLanguage.__munge(execScriptFile)])
        args.extend(BashLanguage.__munge(os.path.abspath(a)) for a in spec.args)

        return (realScriptFile, execScriptFile, args)

    @staticmethod
    def mangleFingerprints(scriptFragments, env):
        # join the script fragments first
        script = joinScripts(scriptFragments, BashLanguage.glue)

        # do not add preamble for empty scripts
        if not script: return ""

        # Add snippets as they match and a default settings preamble
        ret = [script]
        for n,s in BASH_FINGERPRINT_SNIPPETS:
            if n in script: ret.append(s)
        ret.extend(["set -o errexit", "set -o nounset", "set -o pipefail"])
        for n,v in sorted(env.items()):
            ret.append("export {}={}".format(n, quote(v)))
        return "\n".join(reversed(ret))

    @staticmethod
    def setupFingerprint(spec, env):
        env["BOB_CWD"] = BashLanguage.__munge(env["BOB_CWD"])
        return ["bash", "-x", "-c", spec.fingerprintScript]


class PwshResolver(IncludeResolver):
    def __init__(self, fileLoader, baseDir, origText, sourceName, varBase):
        super().__init__(fileLoader, baseDir, origText, sourceName, varBase)
        self.prolog = []
        self.count = 0

    def _includeFile(self, content):
        var = "$_{}{}".format(self.varBase, self.count)
        self.count += 1
        self.prolog.extend([
            "{VAR} = (New-TemporaryFile).FullName".format(VAR=var),
            "$_BOB_TMP_CLEANUP += {VAR}".format(VAR=var),
            "[Convert]::FromBase64String(@'"])
        self.prolog.extend(sliceString(b64encode(content).decode("ascii"), 76))
        self.prolog.append("'@) | Set-Content {VAR} -AsByteStream".format(VAR=var))
        return var

    def _includeLiteral(self, content):
        return quotePwsh(content.decode('utf8'))

    def _resolveContent(self, result):
        return "\n".join(self.prolog + [result])


class PwshLanguage:
    index = ScriptLanguage.PWSH
    glue = "\ncd $Env:BOB_CWD\n"
    Resolver = PwshResolver

    HELPERS = dedent("""\
        function Check-Command {
            param (
                [scriptblock]$ScriptBlock,
                [string]$ErrorAction = $ErrorActionPreference
            )
            & @ScriptBlock
            if (($lastexitcode -ne 0) -and $ErrorAction -eq "Stop") {
                exit $lastexitcode
            }
        }
        """)

    @staticmethod
    def __formatSetup(spec):
        pathSep = ";" if sys.platform == "win32" else ":"
        env = { key: escapePwsh(value) for (key, value) in spec.env.items() }
        env.update({
            "PATH": pathSep.join(
                [escapePwsh(os.path.abspath(p)) for p in spec.paths] +
                (["$Env:PATH"] if not spec.hasSandbox else
                 [escapePwsh(p) for p in spec.sandboxPaths])
            ),
            "LD_LIBRARY_PATH": pathSep.join(
                escapePwsh(os.path.abspath(p)) for p in spec.libraryPaths
            ),
            "BOB_CWD": escapePwsh(os.path.abspath(spec.workspaceExecPath)),
        })

        ret = [
            "# Special Bob array variables:",
            "$BOB_ALL_PATHS=@{{ {} }}".format("; ".join(sorted(
                [ '{} = "{}"'.format(quotePwsh(name), escapePwsh(os.path.abspath(path)))
                    for name,path in spec.allPaths ] ))),
            "$BOB_DEP_PATHS=@{{ {} }}".format("; ".join(sorted(
                [ '{} = "{}"'.format(quotePwsh(name), escapePwsh(os.path.abspath(path)))
                    for name,path in spec.depPaths ] ))),
            "$BOB_TOOL_PATHS=@{{ {} }}".format("; ".join(sorted(
                [ '{} = "{}"'.format(quotePwsh(name), escapePwsh(os.path.abspath(path)))
                    for name,path in spec.toolPaths ] ))),
            "",
            "# Environment:",
            "\n".join('$Env:{}="{}"'.format(k, v) for (k,v) in sorted(env.items())),
            "",
            "# Convenience helpers",
            PwshLanguage.HELPERS,
        ]
        return "\n".join(ret)

    @staticmethod
    def __formatScript(spec, trace):
        envFile = "/bob/env" if spec.hasSandbox else os.path.abspath(spec.envFile)
        ret = [
            PwshLanguage.__formatSetup(spec),
            "",
            dedent("""\
                # Setup
                $ret = [ordered]@{{
                    "Env" = [ordered]@{{}};
                    "Vars" = [ordered]@{{}}
                }}
                foreach ($i in (Get-Variable * | Sort-Object -Property Name)) {{
                    $ret["Vars"][$i.Name] = $i.Value
                }}
                foreach ($i in (Get-Item Env:* | Sort-Object -Property Name)) {{
                    $ret["Env"][$i.Name] = $i.Value
                }}
                $ret["Vars"].Remove("ret")
                ConvertTo-Json $ret -Compress -Depth 2 > {ENV_FILE}
                cd $Env:BOB_CWD
                """.format(ENV_FILE=quotePwsh(envFile))),
            dedent("""\
                # Error handling
                $ErrorActionPreference="Stop"
                Set-PSDebug -Strict
                """),
            "",
            dedent("""\
                try {
                    $_BOB_TMP_CLEANUP = @()
                """),
            "# BEGIN BUILD SCRIPT",
            spec.script,
            "# END BUILD SCRIPT",
            dedent("""\
                } finally {
                    foreach($f in $_BOB_TMP_CLEANUP) {
                        Remove-Item $f -Force
                    }
                }"""),
        ]
        return "\n".join(ret)

    @staticmethod
    def __scriptFilePaths(spec, tmpDir):
        if spec.hasSandbox:
            execScriptFile = "/.script.ps1"
            realScriptFile = (spec.scriptHint or os.path.join(tmpDir, ".script")) + ".ps1"
        else:
            execScriptFile = (spec.scriptHint or os.path.join(tmpDir, "script")) + ".ps1"
            realScriptFile = execScriptFile
        return (os.path.abspath(realScriptFile), os.path.abspath(execScriptFile))

    @staticmethod
    def setupShell(spec, tmpDir, keepEnv):
        realScriptFile, execScriptFile = PwshLanguage.__scriptFilePaths(spec, tmpDir)
        with open(realScriptFile, "w") as f:
            f.write(PwshLanguage.__formatSetup(spec))

        interpreter = "powershell" if isWindows() else "pwsh"
        args = [interpreter, "-ExecutionPolicy", "Bypass", "-NoExit", "-File",
            execScriptFile]
        args.extend(os.path.abspath(a) for a in spec.args)

        return (realScriptFile, execScriptFile, args)

    @staticmethod
    def setupCall(spec, tmpDir, keepEnv, trace):
        realScriptFile, execScriptFile = PwshLanguage.__scriptFilePaths(spec, tmpDir)
        with open(realScriptFile, "w") as f:
            f.write(PwshLanguage.__formatScript(spec, trace))

        interpreter = "powershell" if isWindows() else "pwsh"
        args = [interpreter, "-ExecutionPolicy", "Bypass", "-File", execScriptFile]
        args.extend(os.path.abspath(a) for a in spec.args)

        return (realScriptFile, execScriptFile, args)

    @staticmethod
    def mangleFingerprints(scriptFragments, env):
        # join the script fragments first
        script = joinScripts(scriptFragments, PwshLanguage.glue)

        # do not add preamble for empty scripts
        if not script: return ""

        # Add snippets as they match and a default settings preamble
        ret = [script]
        ret.extend(['$ErrorActionPreference="Stop"', 'Set-PSDebug -Strict'])
        for n,v in sorted(env.items()):
            ret.append('$Env:{}="{}"'.format(k, escapePwsh(v)))
        ret.append(PwshLanguage.HELPERS)

        return "\n".join(reversed(ret))

    @staticmethod
    def setupFingerprint(spec, env):
        interpreter = "powershell" if isWindows() else "pwsh"
        return [interpreter, "-c", spec.fingerprintScript]


LANG = {
    ScriptLanguage.BASH : BashLanguage,
    ScriptLanguage.PWSH : PwshLanguage,
}

def getLanguage(language):
    return LANG[language]


class StepSpec:

    @classmethod
    def fromStep(cls, step, envFile, envWhiteList, logFile=None, isJenkins=False,
                 scriptHint=None):
        self = cls()
        scriptLanguage = step.getPackage().getRecipe().scriptLanguage
        self.__data = d = {
            'envFile' : envFile,
            'envWhiteList' : sorted(envWhiteList),
            'logFile' : logFile,
            'isJenkins' : isJenkins,
            'scriptHint' : scriptHint,
            'vsn' : asHexStr(BOB_INPUT_HASH),
            'language' : scriptLanguage.index.value,
            'env' : dict(step.getEnv()),
            'paths' : step.getPaths(),
            'libraryPaths' : step.getLibraryPaths(),
            'workspace' : (step.getWorkspacePath(), step.getExecPath()),
            'args' : [ a.getExecPath(step) for a in step.getArguments() ],
            'allPaths' : sorted([
                (a.getPackage().getName(), a.getExecPath(step))
                for a in step.getAllDepSteps()
            ]),
            'depPaths' : sorted([
                (a.getPackage().getName(), a.getExecPath(step))
                for a in step.getArguments() if a.isValid()
            ]),
            'toolPaths' : sorted([
                (n, os.path.join(t.getStep().getExecPath(step), t.getPath()))
                for (n,t) in step.getTools().items()
            ]),
        }

        if step.isCheckoutStep():
            d['clean'] = False
        elif step.isPackageStep():
            d['clean'] = True
        else:
            d['clean'] = None

        # fetch sandbox if configured
        if step.getSandbox() is not None:
            d['sandbox'] = s = {
                'root' : step.getSandbox().getStep().getWorkspacePath(),
                'paths' : step.getSandbox().getPaths(),
                'hostMounts' : step.getSandbox().getMounts(),
                'netAccess' : step.hasNetAccess(),
                'depMounts' : [
                    (dep.getWorkspacePath(), dep.getExecPath(step))
                    for dep in step.getAllDepSteps() if dep.isValid()
                ],
            }

            # special handling to mount all previous steps of current package
            extra = step
            while extra.isValid() and len(extra.getArguments()) > 0:
                extra = extra.getArguments()[0]
                if extra.isValid():
                    s['depMounts'].append((extra.getWorkspacePath(), extra.getExecPath(step)))

        d['preRunCmds'] = step.getJenkinsPreRunCmds() if isJenkins else step.getPreRunCmds()
        d['script'] = step.getJenkinsScript() if isJenkins else step.getScript()
        d['postRunCmds'] = step.getJenkinsPostRunCmds() if isJenkins else step.getPostRunCmds()
        d['fingerprintScript'] = step._getFingerprintScript()

        return self

    @classmethod
    def fromFile(cls, f):
        d = json.load(f)
        if d.get('vsn') != asHexStr(BOB_INPUT_HASH):
            raise ParseError("The spec file was created by a different Bob version and is incompatible",
                help="Please re-run this step via bob to fix this error.")
        self = cls()
        self.__data = d
        return self

    def toFile(self, f):
        json.dump(self.__data, f, indent="\t", sort_keys=True)

    def toString(self):
        return json.dumps(self.__data, indent="\t", sort_keys=True)

    @property
    def hasSandbox(self):
        return 'sandbox' in self.__data

    @property
    def language(self):
        return getLanguage(ScriptLanguage(self.__data['language']))

    @property
    def workspaceWorkspacePath(self):
        return self.__data['workspace'][0]

    @property
    def workspaceExecPath(self):
        return self.__data['workspace'][1]

    @property
    def args(self):
        return self.__data['args']

    @property
    def env(self):
        return self.__data['env']

    @property
    def paths(self):
        return self.__data['paths']

    @property
    def sandboxDepMounts(self):
        return self.__data['sandbox']['depMounts']

    @property
    def sandboxNetAccess(self):
        return self.__data['sandbox']['netAccess']

    @property
    def sandboxRootWorkspace(self):
        return self.__data['sandbox']['root']

    @property
    def sandboxHostMounts(self):
        return self.__data['sandbox']['hostMounts']

    @property
    def sandboxPaths(self):
        return self.__data['sandbox']['paths']

    @property
    def envWhiteList(self):
        return set(self.__data['envWhiteList'])

    @property
    def envFile(self):
        return self.__data['envFile']

    @property
    def isJenkins(self):
        return self.__data['isJenkins']

    @property
    def clean(self):
        return self.__data['clean']

    @property
    def logFile(self):
        return self.__data['logFile']

    @property
    def preRunCmds(self):
        return self.__data['preRunCmds']

    @property
    def script(self):
        return self.__data['script']

    @property
    def fingerprintScript(self):
        return self.__data['fingerprintScript']

    @property
    def postRunCmds(self):
        return self.__data['postRunCmds']

    @property
    def libraryPaths(self):
        return self.__data['libraryPaths']

    @property
    def allPaths(self):
        return self.__data['allPaths']

    @property
    def depPaths(self):
        return self.__data['depPaths']

    @property
    def toolPaths(self):
        return self.__data['toolPaths']

    @property
    def scriptHint(self):
        return self.__data['scriptHint']
